<?php
/**
 * Plugin Name: Import Paid Memberships Pro Members from CSV
 * Plugin URI: http://wordpress.org/plugins/pmpro-import-members-from-csv/
 * Description: Import Users and their metadata from a csv file.
 * Version: 2.50
 * Requires PHP: 5.4
 * Author: <a href="https://eighty20results.com/thomas-sjolshagen/">Thomas Sjolshagen <thomas@eighty20results.com></a>
 * License: GPL2
 * Text Domain: pmpro-import-members-from-csv
 * Domain Path: languages/
 */

/**
 * Copyright 2017-2018 - Thomas Sjolshagen (https://eighty20results.com/thomas-sjolshagen)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @credit http://wordpress.org/plugins/import-users-from-csv/ - Ulich Sossou -  https://github.com/sorich87
 * @credit https://github.com/strangerstudios/pmpro-import-users-from-csv - Jason Coleman - https://github.com/ideadude
 */

namespace PMPRO\Addons;

if ( ! defined( 'PMP_IM_CSV_DELIMITER' ) ) {
	define( 'PMP_IM_CSV_DELIMITER', ',' );
}
if ( ! defined( 'PMP_IM_CSV_ESCAPE' ) ) {
	define( 'PMP_IM_CSV_ESCAPE', '\\' );
}
if ( ! defined( 'PMP_IM_CSV_ENCLOSURE' ) ) {
	define( 'PMP_IM_CSV_ENCLOSURE', '"' );
}

if ( ! defined( 'E20R_IMPORT_VERSION' ) ) {
	define( 'E20R_IMPORT_VERSION', '2.50' );
}

class Import_Members_From_CSV {
	
	/**
	 * Instance of this class
	 *
	 * @var null|Import_Members_From_CSV $instance
	 */
	private static $instance = null;
	
	/**
	 * Path to error log file
	 *
	 * @var string $logfile_path
	 */
	private $logfile_path = '';
	
	/**
	 * URI for error log
	 *
	 * @var string $logfile_url
	 */
	private $logfile_url = '';
	
	/**
	 * List of Membership import fields
	 *
	 * @var array|null $pmpro_fields
	 */
	private $pmpro_fields = null;
	
	/**
	 * Name/path of CSV import file
	 *
	 * @var null|string $filename
	 */
	private $filename = null;
	
	/**
	 * Update existing user data?
	 *
	 * @var bool $update_users
	 */
	private $update_users = false;
	
	/**
	 * Set the password nag message when user logs in for the first time?
	 *
	 * @var bool $password_nag
	 */
	private $password_nag = false;
	
	/**
	 * @var bool $password_hashing_disabled - Password is supplied in import file as an encrypted string
	 */
	private $password_hashing_disabled = false;
	
	/**
	 * Should we deactivate old membership levels for the user that
	 * match the record being imported?
	 *
	 * @var bool $deactivate_old_memberships
	 */
	private $deactivate_old_memberships = false;
	
	/**
	 * Do we send the imported user the "new WordPress Account" notice?
	 *
	 * @var bool $new_user_notification
	 */
	private $new_user_notification = false;
	
	/**
	 * Do we send a welcome message to the member if they're imported as an active member to the site
	 *
	 * @var bool $new_member_notification
	 */
	private $new_member_notification = false;
	
	/**
	 * Do we use the imported_member.html template and welcome the imported member?
	 *
	 * @var bool
	 */
	private $send_welcome_email = false;
	
	/**
	 * Do we suppress the "your password was changed" message to the (new/updated) user?
	 *
	 * @var bool
	 */
	private $suppress_pwdmsg = false;
	
	/**
	 * Do we include the admin in the New Member notification email on import?
	 *
	 * @var bool
	 */
	private $admin_new_user_notification = false;
	
	/**
	 * The ID of the multisite to import the user data to/for
	 *
	 * @var null|int $site_id
	 */
	private $site_id = null;
	
	/**
	 * Import the CSV file as a "background" process (i.e. with a JavaScript loop)
	 *
	 * @var bool $background_import
	 */
	private $background_import = false;
	
	/**
	 * Add a PMPro Order record for the imported user (assumes there's either/both initial_payment and billing_amount +
	 * cycle info)
	 *
	 * @var bool $create_order
	 */
	private $create_order = false;
	
	/**
	 * Number of records to import per transaction
	 *
	 * @var int $per_partial
	 */
	private $per_partial = 30;
	
	/**
	 * @var null|\SplFileObject
	 */
	private $file_object = null;
	
	/**
	 * @var null|string
	 */
	private $welcome_mail_warning = null;
	
	/**
	 * Errors to include on the front-end of the import page
	 *
	 * @var array
	 */
	private $display_errors = array();
	
	/**
	 * Import_Members_From_CSV constructor.
	 */
	private function __construct() {
		
		// Set the error log info
		$upload_dir         = wp_upload_dir();
		$this->logfile_path = trailingslashit( $upload_dir['basedir'] ) . 'pmp_im_errors.log';
		$this->logfile_url  = trailingslashit( $upload_dir['baseurl'] ) . 'pmp_im_errors.log';
		
		/*
		$this->pmpro_fields = array(
			"membership_id"                          => null,
			"membership_code_id"                     => null,
			"membership_discount_code"               => null,
			"membership_initial_payment"             => null,
			"membership_billing_amount"              => null,
			"membership_cycle_number"                => null,
			"membership_cycle_period"                => null,
			"membership_billing_limit"               => null,
			"membership_trial_amount"                => null,
			"membership_trial_limit"                 => null,
			"membership_status"                      => 'active',
			"membership_startdate"                   => null,
			"membership_enddate"                     => null,
			"membership_subscription_transaction_id" => null,
			"membership_payment_transaction_id"      => null,
			"membership_gateway"                     => null,
			"membership_affiliate_id"                => null,
			"membership_timestamp"                   => null,
			"membership_subtotal"                    => null,
		);
		*/
		
		/**
		 * Configure fields for order/member import
		 *
		 * @since v2.20 - ENHANCEMENT: Support most of the fields in the pmpro_memberships_users and pmpro_membership_orders tables
		 */
		$this->pmpro_fields = $this->get_table_info( 'pmpro_memberships_users' ) + $this->get_table_info( 'pmpro_membership_orders' );
		
		$this->pmpro_fields['membership_status'] = 'active';
	}
	
	/**
	 * Fetch all columns for the specified table name.
	 *
	 * @param string $table_name
	 *
	 * @return array
	 *
	 * @since v2.20 - ENHANCEMENT: Extract data dynamically from PMPro's custom membership users/orders tables
	 */
	private function get_table_info( $table_name = 'pmpro_memberships_users' ) {
		
		global $wpdb;
		
		$columns    = $wpdb->get_results( sprintf( "SHOW COLUMNS FROM %s", $wpdb->get_blog_prefix() . "{$table_name}" ) );
		$db_columns = array();
		
		if ( empty( $columns ) ) {
			return array();
		}
		
		
		foreach ( $columns as $col ) {
			
			$prefix = null;
			
			switch ( $col->Field ) {
				// Ignore/skip these.
				case 'id':
				case 'modified':
				case 'session_id':
				case 'accountnumber': // Don't risk it (in case users try to import a full card #)
				case 'code': // The order code (need to ignore)
				case 'billing_name':
					//case 'billing_street':
					//case 'billing_city':
					//case 'billing_state':
					//case 'billing_zip':
					//case 'billing_country':
					//case 'billing_phone':
					break; // Skip
				
				case 'membership_id':
					$db_columns[ $col->Field ] = null;
					break;
				
				case 'user_id':
					$db_columns[ $col->Field ] = null;
					break;
				
				default:
					$db_columns["membership_{$col->Field}"] = null;
			}
		}
		
		return $db_columns;
		
	}
	
	/**
	 * Plugin deactivation hook
	 */
	public static function deactivation() {
		
		delete_option( 'e20r_import_has_donated' );
	}
	
	/**
	 * Initialization
	 *
	 * @since 2.0
	 *
	 **/
	public function load_plugins() {
		
		add_action( 'init', array( self::get_instance(), 'load_i18n' ), 5 );
		add_action( 'init', array( self::get_instance(), 'process_csv' ) );
		
		add_action( 'admin_menu', array( self::get_instance(), 'add_admin_pages' ) );
		add_action( 'admin_notices', array( self::get_instance(), 'display_admin_message' ) );
		
		add_action( 'admin_enqueue_scripts', array( self::get_instance(), 'admin_enqueue_scripts' ) );
		add_action( 'wp_ajax_import_members_from_csv', array(
			self::get_instance(),
			'wp_ajax_import_members_from_csv',
		) );
		add_action( 'wp_ajax_cleanup_csv', array( self::get_instance(), 'wp_ajax_cleanup_csv' ) );
		
		// PMPro specific import functionality
		add_action( 'pmp_im_pre_member_import', array( self::get_instance(), 'pre_member_import' ), 10, 2 );
		add_filter( 'pmp_im_import_usermeta', array( self::get_instance(), 'import_usermeta' ), 10, 2 );
		add_action( 'pmp_im_post_member_import', array( self::get_instance(), 'import_membership_info' ), 10, 2 );
		add_action( 'pmp_im_post_member_import', array( self::get_instance(), 'cleanup' ), 9999, 2 );
		
		add_action( 'admin_bar_menu', array( self::get_instance(), 'load_to_pmpro_menu' ), 1001 );
		
		add_action( 'wp_ajax_e20r_visitor_clicked_donation', array( $this, 'save_donation_ip' ) );
		add_action( 'wp_ajax_nopriv_e20r_visitor_clicked_donation', array( $this, 'save_donation_ip' ) );
		
		// Set URIs in plugin listing to plugin support
		add_filter( 'plugin_row_meta', array( self::get_instance(), 'plugin_row_meta' ), 10, 2 );
		
		// Clear action handler(s) from the Import Users from CSV Integration Add-on for PMPro
		add_action( 'wp_loaded', array( $this, 'removeIUCSVSupport' ), 10 );
		add_action( 'wp_loaded', array( $this, 'add_email_templates' ), 11 );
		add_action( 'wp_mail_failed', array( $this, 'mailFailureHandler' ) );
	}
	
	/**
	 * Return or instantiate class for use
	 *
	 * @return Import_Members_From_CSV
	 */
	public static function get_instance() {
		
		if ( is_null( self::$instance ) ) {
			self::$instance = new self;
		}
		
		return self::$instance;
	}
	
	/**
	 * Add message to /wp-admin/ when having problem(s) with sending wp_mail() message
	 *
	 * @param \WP_Error $error
	 */
	public function mailFailureHandler( $error ) {
		
		$this->add_error_msg(
			sprintf(
				__( 'Unable to send email message from Import operation: %s', 'pmpro-import-members-from-csv' ),
				$error->get_error_message( 'wp_mail_failed' )
			),
			'warning'
		);
	}
	
	/**
	 * Add a error/warning/success/info message to /wp-admin/
	 *
	 * @param string $msg
	 * @param string $type
	 */
	private function add_error_msg( $msg, $type = 'info' ) {
		
		$error_msg = get_option( 'pmp_im_error_msg', array() );
		$skip      = false;
		
		foreach ( $error_msg as $e_key => $msg_info ) {
			if ( isset( $msg_info['message'] ) && $msg_info['message'] == $msg ) {
				$skip = true;
				break;
			}
		}
		
		if ( false === $skip && ! empty( $error_msg ) ) {
			$error_msg = array_merge( $error_msg, array( array( 'type' => $type, 'message' => $msg ) ) );
		} else if ( false === $skip && empty( $error_msg ) ) {
			$error_msg = array( array( 'type' => $type, 'message' => $msg ) );
		}
		
		update_option( 'pmp_im_error_msg', $error_msg, 'no' );
	}
	
	/**
	 * Load translation (glotPress friendly)
	 */
	public function load_i18n() {
		
		load_plugin_textdomain(
			'pmpro-import-members-from-csv',
			false,
			basename( dirname( __FILE__ ) ) . '/languages'
		);
	}
	
	/**
	 * Save IP address of client computer if the user clicks the "donate" button
	 */
	public function save_donation_ip() {
		
		if ( WP_DEBUG ) {
			error_log( "Visitor clicked the 'Donate' button: " . print_r( $_REQUEST, true ) );
		}
		
		check_admin_referer( 'pmp-im-import-members', 'pmp-im-import-members-wpnonce' );
		
		if ( WP_DEBUG ) {
			error_log( "Nonce is good" );
		}
		
		$client_ip    = $this->get_client_ip();
		$do_not_track = apply_filters( 'e20r_import_donation_tracking_disabled', false );
		
		if ( ! empty( $client_ip ) && false === $do_not_track ) {
			
			$donated               = get_option( 'e20r_import_has_donated', array() );
			$donated[ $client_ip ] = current_time( 'timestamp' );
			
			if ( true === update_option( 'e20r_import_has_donated', $donated ) ) {
				if ( WP_DEBUG ) {
					error_log( "Visitor ({$client_ip}) clicked the 'Donate' button" );
				}
				wp_send_json_success();
				exit();
			}
		}
		
		wp_send_json_error();
		exit();
	}
	
	/**
	 * Get the IP address for the client/viewer of the page
	 *
	 * @return null|string
	 */
	private function get_client_ip() {
		
		$client    = @$_SERVER['HTTP_CLIENT_IP'];
		$forward   = @$_SERVER['HTTP_X_FORWARDED_FOR'];
		$remote    = $_SERVER['REMOTE_ADDR'];
		$client_ip = null;
		
		if ( filter_var( $client, FILTER_VALIDATE_IP ) ) {
			$client_ip = $client;
		} else if ( filter_var( $forward, FILTER_VALIDATE_IP ) ) {
			$client_ip = $forward;
		} else {
			$client_ip = $remote;
		}
		
		return $client_ip;
	}
	
	/**
	 * Content of the settings page
	 *
	 * @since 0.1
	 **/
	public function import_page() {
		
		if ( ! current_user_can( 'create_users' ) ) {
			wp_die( __( 'You do not have sufficient permissions to access this page.', 'pmpro-import-members-from-csv' ) );
		}
		
		if ( is_multisite() ) {
			$current_blog_id = get_current_blog_id();
			
			if ( 1 !== $current_blog_id && false === $this->is_pmpro_active() ) { ?>
                <div class="notice notice-error is-dismissable">
					<?php _e( "Paid Memberships Pro is not activated on this sub-site. This import tool will not behave as expected!", "pmpro-import-members-from-csv" ); ?>
                </div> <?php
			}
		}
		
		$all_levels = pmpro_getAllLevels( true, true );
		
		if ( empty( $all_levels ) ) { ?>
            <div class="notice notice-error is-dismissable">
				<?php _e( "This site has no defined membership levels. Without them, this tool can't import any membership data for your users.", "pmpro-import-members-from-csv" ); ?>
            </div> <?php
		} ?>
        <div class="wrap">
		<?php printf( '<div id="e20r-status" %s></div>', ( ! isset( $_REQUEST['import'] ) ? 'style="display: none;"' : 'style="display: inline-block;"' ) ); ?>
        <h2><?php _e( 'Import PMPro members from a CSV file', 'pmpro-import-members-from-csv' ); ?></h2><?php
		
		if ( false === ( $efh = fopen( $this->logfile_path, 'a' ) ) ) {
			
			printf( '<div class="updated"><p><strong>%s</strong></p></div>',
				sprintf(
					__( 'Note: Please make the %s directory writable to see/save the error log.', 'pmpro-import-members-from-csv' ),
					$this->logfile_path
				)
			);
		}
		
		@fwrite( $efh,
			sprintf(
				__( "BEGIN %s\n", 'pmpro-import-members-from-csv' ),
				date( 'Y-m-d H:i:s', current_time( 'timestamp' )
				)
			)
		);
		
		// File was writable and accessible
		fclose( $efh );
		
		if ( isset( $_REQUEST['import'] ) ) {
			$error_log_msg = '';
			
			if ( filesize( $this->logfile_path ) > 0 ) {
				$error_log_msg = sprintf(
					__( ', please %1$scheck the error log%2$s', 'pmpro-import-members-from-csv' ),
					sprintf( '<a href="%s">', esc_url_raw( $this->logfile_url ) ),
					'</a>'
				);
			}
			
			switch ( $_REQUEST['import'] ) {
				case 'file':
					printf( '<div class="error"><p><strong>%s</strong></p></div>', __( 'Error during file upload.', 'pmpro-import-members-from-csv' ) );
					break;
				case 'data':
					printf( '<div class="error"><p><strong>%s</strong></p></div>', __( 'Cannot extract data from uploaded file or no file was uploaded.', 'pmpro-import-members-from-csv' ) );
					break;
				case 'fail':
					printf( '<div class="error"><p><strong>%s</strong></p></div>', sprintf( __( 'No members imported%s.', 'pmpro-import-members-from-csv' ), $error_log_msg ) );
					break;
				case 'errors':
					printf( '<div class="error"><p><strong>%s</strong></p></div>', sprintf( __( 'Some members were successfully imported, but some were not%s.', 'pmpro-import-members-from-csv' ), $error_log_msg ) );
					break;
				case 'success':
					printf( '<div class="updated"><p><strong>%s</strong></p></div>', __( 'Member import was successful.', 'pmpro-import-members-from-csv' ) );
					break;
				default:
			}
			
			if ( isset( $_REQUEST['import'] ) && $_REQUEST['import'] == 'resume' && ! empty( $_REQUEST['filename'] ) ) {
				
				$this->filename = sanitize_file_name( $_REQUEST['filename'] );
				
				//Resetting position option?
				if ( ! empty( $_REQUEST['reset'] ) ) {
					$file = basename( $this->filename );
					delete_option( "pmpcsv_{$file}" );
				} ?>
                <h3><?php _e( 'Importing the file using AJAX (in the background)', 'pmpro-import-members-from-csv' ); ?></h3>
                <p><strong><?php _e( 'IMPORTANT:', 'pmpro-import-members-from-csv' ); ?></strong> <?php printf(
						__( 'Your import is not finished. %1$sClosing this page will stop the import operation%2$s. If the import stops or you have to close your browser, you can navigate to %3$sthis URL%4$s to resume the import operation later.', 'pmpro-import-members-from-csv' ),
						'<strong>',
						'</strong>',
						sprintf( '<a href="%s">', admin_url( 'admin.php' . "?{$_SERVER['QUERY_STRING']}" ) ),
						'</a>'
					); ?>
                </p>

                <p>
                    <a id="pauseimport"
                       href="#"><?php _e( "Click here to pause.", 'pmpro-import-members-from-csv' ); ?></a>
                    <a id="resumeimport" href="#"
                       style="display:none;"><?php _e( "Paused. Click here to resume.", "pmpro-import-members-from-csv" ); ?></a>
                </p>

                <textarea id="importstatus" rows="10"
                          cols="60"><?php _e( 'Loading...', 'pmpro-import-members-from-csv' ); ?></textarea>
                <p class="complete_btn">
                <input type="button" class="button-primary" id="completedImport"
                       value="<?php _e( "Finished", "pmpro-import-members-from-csv" ); ?>" style="display:none;"/>
                </p><?php
			}
		}
		
		if ( empty( $_REQUEST['filename'] ) ) {
			
			$has_donated    = false;
			$current_client = $this->get_client_ip();
			
			/**
			 * Add "donations welcome" button info if the user hasn't recently donated/clicked the button
			 */
			if ( ! empty( $current_client ) ) {
				
				$donated_from = get_option( 'e20r_import_has_donated', array() );
				
				if ( in_array( $current_client, array_keys( $donated_from ) ) ) {
					
					if ( ! empty( $donated_from[ $current_client ] ) ) {
						
						// Get the timestamp when they clicked the 'donate' button
						$when = strtotime(
							date(
								'Y-m-d H:i:s',
								$donated_from[ $current_client ] ) . " +2 month",
							current_time( 'timestamp' )
						);
						
						// Need to clean up
						if ( $when < current_time( 'timestamp' ) ) {
							
							if ( WP_DEBUG ) {
								error_log( "Removing 'donated_from' entry for {$current_client}" );
							}
							
							unset( $donated_from[ $current_client ] );
							update_option( 'e20r_import_has_donated', $donated_from );
						} else {
							$has_donated = true;
						}
					}
				}
			} ?>
            <div class="e20r-donation-button">
				<?php if ( $has_donated === false ) { ?>
                    <p class="e20r-donation-text"><?php _e( "We have a donation button in case you wish to show your support for the continued development and support activities for the plugin.", "pmpro-import-members-from-csv" ); ?></p>
                    <p class="e20r-donation-text"><?php _e( 'Whether you choose to donate or not is obviously up to you. We do hope the plugin has enough value for you to click on the "Donate" button and help us out, but we also get it if you can\'t at this time.', 'pmpro-import-members-from-csv' ); ?></p>
				<?php } ?>
                <p class="e20r-donation-text"><?php printf(
						__(
							'%1$sIf you haven\'t reviewed our plugin, we would %2$sappreciate your honest feedback%3$s!%4$s',
							'pmpro-import-members-from-csv'
						),
						'<strong>',
						sprintf( '<a href="%1$s">', esc_url( 'https://wordpress.org/support/plugin/pmpro-import-members-from-csv/reviews/#new-post' ) ),
						'</a>',
						'</strong>'
					); ?></p>
				<?php if ( $has_donated === false ) { ?>
                    <p class="e20r-donation-text"><?php printf( __( 'Regardless, please accept our sincere %1$sthanks%2$s for choosing our plugin.', 'pmpro-import-members-from-csv' ), '<strong>', '</strong>' ); ?></p>
                    <div class="e20r-center-button">
                        <form action="https://www.paypal.com/cgi-bin/webscr" method="post" target="_top"
                              id="e20r-import-donation">
                            <input type="hidden" name="cmd" value="_s-xclick">
                            <input type="hidden" name="hosted_button_id" value="YR423UJ7AZJFJ">
                            <input type="image" src="https://www.paypalobjects.com/en_US/i/btn/btn_donateCC_LG.gif"
                                   border="0"
                                   name="submit" alt="PayPal - The safer, easier way to pay online!"
                                   id="e20r_donation_button">
                            <img alt="" border="0" src="https://www.paypalobjects.com/en_US/i/scr/pixel.gif" width="1"
                                 height="1">
                        </form>
                    </div>
                    <p class="e20r-donation-text"><?php _e( 'PS: If you click the "Donate" button, this reminder will disappear for a while', 'pmpro-import-members-from-csv' ); ?></p>
				<?php } ?>
            </div>
            <hr/>
            <form method="post" action="" id="e20r-import-form" enctype="multipart/form-data">
				<?php wp_nonce_field( 'pmp-im-import-members', 'pmp-im-import-members-wpnonce' ); ?>
                <table class="form-table">
                    <tr valign="top">
                        <th scope="row"><label
                                    for="members_csv"><?php _e( 'CSV file to load', 'pmpro-import-members-from-csv' ); ?></label>
                        </th>
                        <td>
                            <input type="file" id="members_csv" name="members_csv" value="" class="all-options"
                                   accept=".csv, text/csv"
                                   about="<?php __( 'Select .CSV file to process', 'pmpro-import-members-from-csv' ); ?>"/><br/>
                            <span class="description"><?php printf( __( 'You may want to download and review <a href="%s" target="_blank">the example CSV file</a>.', 'pmpro-import-members-from-csv' ), plugin_dir_url( __FILE__ ) . 'examples/import.csv' ); ?></span>
                        </td>
                    </tr>
                    <tr valign="top">
                        <th scope="row"><?php _e( 'Update user record', 'pmpro-import-members-from-csv' ); ?></th>
                        <td>
                            <fieldset>
                                <legend class="screen-reader-text">
                                    <span><?php _e( 'Users update', 'pmpro-import-members-from-csv' ); ?></span>
                                </legend>
                                <label for="update_users">
                                    <input id="update_users" name="update_users" type="checkbox" value="1"
                                           checked="checked"/>
									<?php _e( "Update, do not add, the user if the username or email already exists (Recommended)", 'pmpro-import-members-from-csv' ); ?>
                                </label>
                            </fieldset>
                        </td>
                    </tr>
                    <tr valign="top">
                        <th scope="row"><?php _e( 'Deactivate existing membership', 'pmpro-import-members-from-csv' ); ?></th>
                        <td>
                            <fieldset>
                                <legend class="screen-reader-text">
                                    <span><?php _e( 'Deactivate existing membership', 'pmpro-import-members-from-csv' ); ?></span>
                                </legend>
                                <label for="deactivate_">
                                    <input id="deactivate_old_memberships" name="deactivate_old_memberships"
                                           type="checkbox" value="1" checked="checked"/>
									<?php _e( "Update the status when importing a user who already has an 'active' membership level (Recommended)", "pmpro-import-members-from-csv" ); ?>
                                </label>
                            </fieldset>
                        </td>
                    </tr>
                    <tr valign="top">
                        <th scope="row"><?php _e( 'Add order', 'pmpro-import-members-from-csv' ); ?></th>
                        <td>
                            <fieldset>
                                <legend class="screen-reader-text">
                                    <span><?php _e( 'Attempt to create PMPro Order record', 'pmpro-import-members-from-csv' ); ?></span>
                                </legend>
                                <label for="update_users">
                                    <input id="create_order" name="create_order" type="checkbox" value="1"/>
									<?php _e( "Attempt to add a PMPro order record when relevant data is included in the import row", 'pmpro-import-members-from-csv' ); ?>
                                </label>
                            </fieldset>
                        </td>
                    </tr>
                    <tr valign="top">
                        <th scope="row"><?php _e( 'Send WordPress notification', 'pmpro-import-members-from-csv' ); ?></th>
                        <td>
                            <fieldset>
                                <legend class="screen-reader-text">
                                    <span><?php _e( 'Send the WordPress new user/updated password notification to the user', 'pmpro-import-members-from-csv' ); ?></span>
                                </legend>
                                <label for="new_user_notification">
                                    <input id="new_user_notification" name="new_user_notification" type="checkbox"
                                           value="1"/>
									<?php _e( 'Send the WordPress new user/updated password notification to the user', 'pmpro-import-members-from-csv' ); ?>
                                </label>
                            </fieldset>
                        </td>
                    </tr>
                    <tr valign="top">
                        <th scope="row"><?php _e( 'Suppress "Password/Email Changed" notification', 'pmpro-import-members-from-csv' ); ?></th>
                        <td>
                            <fieldset>
                                <legend class="screen-reader-text">
                                    <span><?php _e( 'Do not send the "Your password was changed" nor the "Your email address was changed" email messages to the updated user', 'pmpro-import-members-from-csv' ); ?></span>
                                </legend>
                                <label for="suppress_pwdmsg">
                                    <input id="suppress_pwdmsg" name="suppress_pwdmsg" type="checkbox" value="1"/>
									<?php printf( __( '%1$sDo not send%2$s the "Your password was changed" nor the "Your email address was changed" email messages to the updated user', 'pmpro-import-members-from-csv' ), '<strong>', '</strong>' ); ?>
                                </label>
                            </fieldset>
                        </td>
                    </tr>
                    <!-- send_welcome_email -->
                    <tr valign="top">
                        <th scope="row"><?php _e( 'Send WordPress notification to admin', 'pmpro-import-members-from-csv' ); ?></th>
                        <td>
                            <fieldset>
                                <legend class="screen-reader-text">
                                    <span><?php _e( 'Send WordPress new user/updated user notification to admin', 'pmpro-import-members-from-csv' ); ?></span>
                                </legend>
                                <label for="admin_new_user_notification">
                                    <input id="admin_new_user_notification" name="admin_new_user_notification"
                                           type="checkbox"
                                           value="1"/>
									<?php _e( 'Send WordPress new user/updated user notification to admin', 'pmpro-import-members-from-csv' ); ?>
                                </label>
                            </fieldset>
                        </td>
                    </tr>
                    <tr valign="top">
                        <th scope="row"><?php _e( 'Send \'Welcome to the membership\' email', 'pmpro-import-members-from-csv' ); ?></th>
                        <td>
                            <fieldset>
                                <legend class="screen-reader-text">
                                    <span><?php _e( 'Send the imported_member.html welcome email to the imported user', 'pmpro-import-members-from-csv' ); ?></span>
                                </legend>
                                <label for="send_welcome_email">
                                    <input id="send_welcome_email" name="send_welcome_email" type="checkbox" value="1"/>
									<?php _e( 'Send the imported_member.html welcome email to the imported user', 'pmpro-import-members-from-csv' ); ?>
                                </label>
                            </fieldset>
                        </td>
                    </tr>
                    <!--  -->
                    <tr valign="top">
                        <th scope="row"><?php _e( 'Display password nag', 'pmpro-import-members-from-csv' ); ?></th>
                        <td>
                            <fieldset>
                                <legend class="screen-reader-text">
                                    <span><?php _e( 'Password nag', 'pmpro-import-members-from-csv' ); ?></span>
                                </legend>
                                <label for="password_nag">
                                    <input id="password_nag" name="password_nag" type="checkbox" value="1"/>
									<?php _e( 'Show the password nag when the new user(s) log in', 'pmpro-import-members-from-csv' ); ?>
                                </label>
                            </fieldset>
                        </td>
                    </tr>
                    <tr valign="top">
                        <th scope="row"><?php _e( 'Password is already hashed', 'pmpro-import-members-from-csv' ); ?></th>
                        <td>
                            <fieldset>
                                <legend class="screen-reader-text">
                                    <span><?php _e( 'Password is hashed', 'pmpro-import-members-from-csv' ); ?></span>
                                </legend>
                                <label for="password_hashing_disabled">
                                    <input id="password_hashing_disabled" name="password_hashing_disabled"
                                           type="checkbox" value="1"/>
									<?php _e( "The passsword in the .csv file is already hashed and doesn't need to be encrypted by the import process.", 'pmpro-import-members-from-csv' ); ?>
                                </label>
                            </fieldset>
                        </td>
                    </tr>
                    <tr valign="top">
                        <th scope="row"><?php _e( 'Background import', 'pmpro-import-members-from-csv' ); ?></th>
                        <td>
                            <fieldset>
                                <legend class="screen-reader-text">
                                    <span><?php _e( 'Import the .csv file with resume functionality.', 'pmpro-import-members-from-csv' ); ?></span>
                                </legend>
                                <label for="background_import">
                                    <input id="background_import" name="background_import" type="checkbox" value="1"
                                           checked="checked"/>
									<?php _e( 'Use a background process to import all of the records. (Recommeded)', 'pmpro-import-members-from-csv' ); ?>
                                </label>
                            </fieldset>
                        </td>
                    </tr> <?php
					if ( is_multisite() ) {
						
						$site_list = get_sites();
						?>
                        <tr valign="top">
                        <th scope="row"><?php _e( 'Site to import to', 'pmpro-import-members-from-csv' ); ?></th>
                        <td>
                            <fieldset>
                                <legend class="screen-reader-text">
                                    <span><?php _e( 'Select the multisite instance to import these members to.', 'pmpro-import-members-from-csv' ); ?></span>
                                </legend>
                                <select id="site_id" name="site_id">
									<?php foreach ( $site_list as $site ) { ?>
                                        <option
                                        value="<?php esc_attr_e( $site->blog_id ); ?>" <?php selected( $site->blog_id, $current_blog_id ); ?>><?php esc_html_e( $site->blogname ); ?></option><?php
									} ?>
                                </select>
                            </fieldset>
                        </td>
                        </tr><?php
					}
					?>
					<?php do_action( 'pmp_im_import_page_setting_html' ) ?>
                </table>
                <p class="submit">
                    <input type="submit" class="button-primary"
                           value="<?php _e( 'Import', 'pmpro-import-members-from-csv' ); ?>"
                           id="e20r-import-form-submit"/>
                </p>
            </form>
		<?php } ?>
        </div><?php
	}
	
	/**
	 * Is the Paid Memberships Pro plugin active on the current site?
	 *
	 * @return bool
	 */
	private function is_pmpro_active() {
		
		$active_plugins = get_site_option( 'active_plugins' );
		
		return in_array( 'paid-memberships-pro/paid-memberships-pro.php', $active_plugins );
	}
	
	/**
	 * Add to Memberships Menu in WP Back-end (left sidebar)
	 *
	 * @since 2.1
	 **/
	public function add_admin_pages() {
		
		add_submenu_page(
			'pmpro-membershiplevels',
			__( 'Import Members', 'pmpro-import-members-from-csv' ),
			__( 'Import Members', 'pmpro-import-members-from-csv' ),
			'create_users',
			'pmpro-import-members-from-csv',
			array( self::get_instance(), 'import_page' )
		);
	}
	
	/**
	 * Add Import Members to the PMPro Members drop-down menu
	 * @since 2.1
	 */
	public function load_to_pmpro_menu() {
		
		global $wp_admin_bar;
		
		if ( current_user_can( 'create_users' ) ) {
			$wp_admin_bar->add_menu( array(
					'id'     => 'pmpro-import-members-from-csv',
					'parent' => 'paid-memberships-pro',
					'title'  => __( 'Import Members', 'pmpro-import-members-from-csv' ),
					'href'   => add_query_arg( 'page', 'pmpro-import-members-from-csv', get_admin_url( null, 'admin.php' ) ),
				)
			);
		}
	}
	
	/**
	 * Add admin JS
	 *
	 * @param string $hook
	 *
	 * @since 1.0
	 **/
	public function admin_enqueue_scripts( $hook ) {
		
		if ( ! isset( $_REQUEST['page'] ) || $_REQUEST['page'] != 'pmpro-import-members-from-csv' ) {
			return;
		}
		
		$this->load_settings();
		
		/**
		 * Calculate the max timeout for the AJAX calls. Gets padded with a 20% bonus
		 */
		$max_run_time  = (
			apply_filters( 'pmp_im_import_time_per_record', 3 ) *
			apply_filters( 'pmp_im_import_records_per_scan', $this->per_partial )
		);
		$timeout_value = ceil( $max_run_time * 1.2 );
		
		if ( WP_DEBUG ) {
			error_log( "Setting JavaScript timeout for import operations to {$timeout_value} seconds" );
		}
		
		wp_enqueue_style( 'pmpro-import-members-from-csv', plugins_url( 'css/pmpro-import-members-from-csv.css', __FILE__ ), null, E20R_IMPORT_VERSION );
		wp_register_script( 'pmpro-import-members-from-csv', plugins_url( 'javascript/pmpro-import-members-from-csv.js', __FILE__ ), array( 'jquery' ), E20R_IMPORT_VERSION );
		
		wp_localize_script( 'pmpro-import-members-from-csv', 'pmp_im_settings',
			apply_filters( 'pmp_im_import_js_settings', array(
					'timeout'                     => $timeout_value,
					'background_import'           => intval( $this->background_import ),
					'filename'                    => $this->filename,
					'update_users'                => intval( $this->update_users ),
					'deactivate_old_memberships'  => intval( $this->deactivate_old_memberships ),
					'new_user_notification'       => intval( $this->new_user_notification ),
					'create_order'                => intval( $this->create_order ),
					'admin_new_user_notification' => intval( $this->admin_new_user_notification ),
					'send_welcome_email'          => intval( $this->send_welcome_email ),
					'suppress_pwdmsg'             => intval( $this->suppress_pwdmsg ),
					'password_hashing_disabled'   => intval( $this->password_hashing_disabled ),
					'password_nag'                => intval( $this->password_nag ),
					'per_partial'                 => intval( $this->per_partial ),
					'site_id'                     => intval( $this->site_id ),
					'admin_page'                  => add_query_arg( 'page', 'pmpro-import-members-from-csv', admin_url( 'admin.php' ) ),
					'import'                      => isset( $_REQUEST['import'] ) ? sanitize_text_field( $_REQUEST['import'] ) : null,
					'lang'                        => array(
						'whitespace_in_filename' => __( 'Error: Your file name contains one or more whitespace characters. Please rename the file and remove any whitespace characters from the file name.', 'pmpro-import-members-from-csv' ),
						'pausing'                => __( 'Pausing. You may see one more update here as we clean up.', 'pmpro-import-members-from-csv' ),
						'resuming'               => __( 'Resuming...', 'pmpro-import-members-from-csv' ),
						'loaded'                 => __( 'JavaScript Loaded.', 'pmpro-import-members-from-csv' ),
						'done'                   => __( 'Done!', 'pmpro-import-members-from-csv' ),
						'alert_msg'              => __( 'Error with import. Close to reload the admin page.', 'pmpro-import-members-from-csv' ),
						'error'                  => __( 'Error with import. Close to refresh the admin page.', 'pmpro-import-members-from-csv' ),
						'excel_info'             => sprintf( __( 'If you use Microsoft Excel(tm) to view/edit your .CSV files, may we suggest you %1$stry using Google Sheets instead%2$s? Using Google Sheets may reduce/eliminate issues with date formats!', 'pmpro-import-members-from-csv' ), sprintf( '<a href="%s" target="_blank" title="%s">', 'http://docs.google.com/spreadsheets', __( 'To Google Sheets', 'pmpro-import-members-from-csv' ) ), '</a>' ),
					),
					'display_errors'              => ( ! empty( $this->display_errors ) ? $this->display_errors : null ),
				)
			)
		);
		
		wp_enqueue_script( 'pmpro-import-members-from-csv' );
	}
	
	/**
	 * Load/configure settings from $_REQUEST array (if available)
	 */
	public function load_settings() {
		
		
		if ( true === $this->is_configured() ) {
			return;
		}
		
		$this->filename = isset( $_FILES['members_csv']['tmp_name'] ) ? $_FILES['members_csv']['tmp_name'] : $this->filename;
		$this->filename = isset( $_REQUEST['filename'] ) ? sanitize_file_name( $_REQUEST['filename'] ) : $this->filename;
		
		if ( WP_DEBUG ) {
			error_log( "Setting file name to {$this->filename}" );
		}
		
		$this->update_users = ! empty( $_REQUEST['update_users'] ) ? ( 1 === intval( $_REQUEST['update_users'] ) ) : $this->update_users;
		
		if ( WP_DEBUG ) {
			error_log( "Settings users update to: " . ( $this->update_users ? "True" : "False" ) );
		}
		
		$this->background_import           = ! empty( $_REQUEST['background_import'] ) ? ( 1 === intval( $_REQUEST['background_import'] ) ) : $this->background_import;
		$this->deactivate_old_memberships  = ! empty( $_REQUEST['deactivate_old_memberships'] ) ? ( 1 === intval( $_REQUEST['deactivate_old_memberships'] ) ) : $this->deactivate_old_memberships;
		$this->create_order                = ! empty( $_REQUEST['create_order'] ) ? ( 1 === intval( $_REQUEST['create_order'] ) ) : $this->create_order;
		$this->password_nag                = ! empty( $_REQUEST['password_nag'] ) ? ( 1 === intval( $_REQUEST['password_nag'] ) ) : $this->password_nag;
		$this->password_hashing_disabled   = ! empty( $_REQUEST['password_hashing_disabled'] ) ? ( 1 === intval( $_REQUEST['password_hashing_disabled'] ) ) : $this->password_hashing_disabled;
		$this->new_user_notification       = ! empty( $_REQUEST['new_user_notification'] ) ? ( 1 === intval( $_REQUEST['new_user_notification'] ) ) : $this->new_user_notification;
		$this->suppress_pwdmsg             = ! empty( $_REQUEST['suppress_pwdmsg'] ) ? ( 1 === intval( $_REQUEST['suppress_pwdmsg'] ) ) : $this->suppress_pwdmsg;
		$this->admin_new_user_notification = ! empty( $_REQUEST['admin_new_user_notification'] ) ? 1 === intval( $_REQUEST['admin_new_user_notification'] ) : $this->admin_new_user_notification;
		$this->send_welcome_email          = ! empty( $_REQUEST['send_welcome_email'] ) ? 1 === intval( $_REQUEST['send_welcome_email'] ) : $this->send_welcome_email;
		$this->new_member_notification     = ! empty( $_REQUEST['new_member_notification'] ) ? ( 1 === intval( $_REQUEST['new_member_notification'] ) ) : (bool) $this->new_member_notification;
		$this->per_partial                 = ! empty( $_REQUEST['per_partial'] ) ? intval( $_REQUEST['per_partial'] ) : $this->per_partial;
		$this->site_id                     = ! empty( $_REQUEST['site_id'] ) ? intval( $_REQUEST['site_id'] ) : $this->site_id;
		
		if ( WP_DEBUG ) {
			error_log( "Do we suppress the changed password email? " . ( $this->suppress_pwdmsg ? 'Yes' : 'No' ) );
		}
		
		// Calculate # of records to import per iteration
		$this->calculate_max_records( $this->new_member_notification, (bool) $this->create_order );
		
		$this->per_partial = apply_filters( 'pmp_im_import_records_per_scan', $this->per_partial );
	}
	
	/**
	 * Is the class configured (Request variables read to variables) already?
	 *
	 * @return bool
	 */
	private function is_configured() {
		
		return ! empty( $this->filename );
	}
	
	/**
	 * Calculate the # of records we'll attempt to import per iteration
	 *
	 * @param bool $include_welcome_message
	 * @param bool $include_order_link
	 */
	private function calculate_max_records( $include_welcome_message = false, $include_order_link = false ) {
		
		/**
		 * Calculate the # of records to import per operation when running in background mode
		 */
		$max_exec_time   = intval( floor( intval( ini_get( 'max_execution_time' ) ) * 0.80 ) );
		$per_record_time = apply_filters( 'pmp_im_import_time_per_record', 1.5 );
		
		// Add time for creating order(s) (reduces # of records per iteration)
		if ( true === $include_order_link ) {
			$per_record_time += apply_filters( 'pmp_im_order_link_time', 1 );
		}
		
		// Add time for sending the welcome message (reduces # of records per iteration)
		if ( true === $include_welcome_message ) {
			$per_record_time += apply_filters( 'pmp_im_weclome_msg_time', 4 );
		}
		
		
		if ( ! empty( $max_exec_time ) && ( is_numeric( $per_record_time ) ) ) {
			$this->per_partial = round( ceil( $max_exec_time / (float) $per_record_time ), 0 );
		}
		
		if ( $this->per_partial > 60 ) {
			$this->per_partial = 60;
		}
		
		if ( WP_DEBUG ) {
			error_log( "Will allow up to {$this->per_partial} records per iteration, using up to {$per_record_time} seconds per record..." );
		}
	}
	
	/**
	 * Add admin notice to WP Admin backedn
	 */
	public function display_admin_message() {
		
		$error_msgs = get_option( 'pmp_im_error_msg', array() );
		
		if ( ! empty( $error_msgs ) && is_admin() ) {
			
			foreach ( $error_msgs as $key => $msg_info ) { ?>
            <div class="notice notice-<?php esc_attr_e( $msg_info['type'] ); ?> is-dismissible">
                <p><strong><?php echo $msg_info['message']; ?></strong></p>
                <button type="button" class="notice-dismiss">
                    <span class="screen-reader-text"><?php __( 'Dismiss this message.', 'pmpro-import-members-from-csv' ); ?></span>
                </button>
                </div><?php
			}
			
			// Clear error/warning/notice/success messages
			delete_option( 'pmp_im_error_msg' );
		}
	}
	
	/**
	 * Process content of CSV file
	 *
	 * @since 1.0
	 **/
	public function process_csv() {
		
		if ( isset( $_REQUEST['pmp-im-import-members-wpnonce'] ) &&
		     ( ! isset( $_REQUEST['action'] ) ||
		       ( isset( $_REQUEST['action'] ) && 'import_members_from_csv' == $_REQUEST['action'] )
		     ) && ! ( defined( 'DOING_AJAX' ) && DOING_AJAX !== true )
		) {
			
			if ( WP_DEBUG ) {
				error_log( "Processing AJAX request" );
			}
			
			check_admin_referer( 'pmp-im-import-members', 'pmp-im-import-members-wpnonce' );
			
			// Setup settings variables
			$this->load_settings();
			$errors = array();
			
			if ( isset( $_FILES['members_csv']['tmp_name'] ) ) {
				
				//use AJAX?
				if ( true === $this->background_import ) {
					
					if ( WP_DEBUG ) {
						error_log( "Background processing for import" );
					}
					//check for a imports directory in wp-content
					$upload_dir = wp_upload_dir();
					$import_dir = $upload_dir['basedir'] . "/e20r_imports/";
					
					//create the dir and subdir if needed
					if ( ! is_dir( $import_dir ) ) {
						if ( false === wp_mkdir_p( $import_dir ) ) {
							
							$this->add_error_msg( sprintf( __( "Unable to create directory on your server: %s", 'pmpro-import-members-from-csv' ), $import_dir ), 'error' );
							wp_redirect( add_query_arg( 'import', 'fail', wp_get_referer() ) );
							exit();
						}
					}
					
					//figure out filename
					$this->filename = $_FILES['members_csv']['name'];
					$file_arr       = explode( '.', $this->filename );
					$filetype       = $file_arr[ ( count( $file_arr ) - 1 ) ];
					
					$count = 0;
					
					if ( empty( $this->filename ) ) {
						
						$this->add_error_msg( __( 'CSV file not selected. Nothing to import!', 'import-members-from-csv' ), 'error' );
						wp_redirect( add_query_arg( 'import', 'fail', wp_get_referer() ) );
						exit();
					}
					
					while ( file_exists( "{$import_dir}{$this->filename}" ) ) {
						
						if ( ! empty( $count ) ) {
							$this->filename = $this->str_lreplace( "-{$count}.{$filetype}", "-" . strval( $count + 1 ) . ".{$filetype}", $this->filename );
						} else {
							$this->filename = $this->str_lreplace( ".{$filetype}", "-1.{$filetype}", $this->filename );
						}
						
						$count ++;
						
						//let's not expect more than 50 files with the same name
						if ( $count > 50 ) {
							$this->add_error_msg( sprintf( __( "Error uploading file! Too many files with the same name. Please clean out the %s directory on your server.", "pmpro-import-members-from-csv" ), $import_dir ), 'error' );
							wp_redirect( add_query_arg( 'import', 'fail', wp_get_referer() ) );
							exit();
						}
					}
					
					//save file
					if ( false !== strpos( $_FILES['members_csv']['tmp_name'], $upload_dir['basedir'] ) ) {
						
						//was uploaded and saved to $_SESSION
						rename( $_FILES['members_csv']['tmp_name'], "{$import_dir}{$this->filename}" );
					} else {
						//it was just uploaded
						move_uploaded_file( $_FILES['members_csv']['tmp_name'], "{$import_dir}{$this->filename}" );
					}
					
					//redurect to the page to run AJAX
					$url = add_query_arg(
						array(
							'page'                        => 'pmpro-import-members-from-csv',
							'import'                      => 'resume',
							'filename'                    => $this->filename,
							'background_import'           => true,
							'update_users'                => $this->update_users,
							'password_nag'                => $this->password_nag,
							'password_hashing_disabled'   => $this->password_hashing_disabled,
							'new_user_notification'       => $this->new_user_notification,
							'suppress_pwdmsg'             => $this->suppress_pwdmsg,
							'admin_new_user_notification' => $this->admin_new_user_notification,
							'send_welcome_email'          => $this->send_welcome_email,
							'deactivate_old_memberships'  => $this->deactivate_old_memberships,
							'create_order'                => $this->create_order,
							'partial'                     => true,
							'per_partial'                 => $this->per_partial,
							'site_id'                     => $this->site_id,
						),
						admin_url( 'admin.php' )
					);
					
					wp_redirect( $url );
					exit;
					
				} else {
					
					$results = $this->import_csv( $this->filename, array(
						'filename'                    => $this->filename,
						'background_import'           => true,
						'update_users'                => $this->update_users,
						'password_nag'                => $this->password_nag,
						'password_hashing_disabled'   => $this->password_hashing_disabled,
						'new_user_notification'       => $this->new_user_notification,
						'admin_new_user_notification' => $this->admin_new_user_notification,
						'suppress_pwdmsg'             => $this->suppress_pwdmsg,
						'send_welcome_email'          => $this->send_welcome_email,
						'deactivate_old_memberships'  => $this->deactivate_old_memberships,
						'create_order'                => $this->create_order,
						'partial'                     => false,
						'per_partial'                 => apply_filters( 'pmp_im_import_records_per_scan', 30 ),
						'site_id'                     => $this->site_id,
					) );
					
					// No users imported?
					if ( ! $results['user_ids'] ) {
						wp_redirect( add_query_arg( 'import', 'fail', wp_get_referer() ) );
						
						// Some users imported?
					} else if ( $results['errors'] ) {
						wp_redirect( add_query_arg( 'import', 'errors', wp_get_referer() ) );
						
						// All users imported? :D
					} else {
						wp_redirect( add_query_arg( 'import', 'success', wp_get_referer() ) );
					}
					
					exit;
				}
			}
			
			wp_redirect( add_query_arg( 'import', 'file', wp_get_referer() ) );
			exit;
		}
	}
	
	/**
	 * Replace leftmost instance of string
	 *
	 * @param string $search
	 * @param string $replace
	 * @param string $subject
	 *
	 * @return string
	 */
	public function str_lreplace( $search, $replace, $subject ) {
		
		$pos = strrpos( $subject, $search );
		
		if ( $pos !== false ) {
			$subject = substr_replace( $subject, $replace, $pos, strlen( $search ) );
		}
		
		return $subject;
	}
	
	/**
	 * Import a csv file
	 *
	 * @param array $args
	 *
	 * @return array
	 * @since 0.5
	 */
	public function import_csv( $filename, $args ) {
		
		if ( is_multisite() ) {
			$current_blog_id = get_current_blog_id();
		}
		
		$errors   = array();
		$warnings = array();
		$user_ids = array();
		$headers  = array();
		
		$defaults = array(
			'filename'                    => null,
			'password_nag'                => false,
			'background_import'           => false,
			'new_user_notification'       => false,
			'admin_new_user_notification' => false,
			'send_welcome_email'          => false,
			'suppress_pwdmsg'             => false,
			'password_hashing_disabled'   => false,
			'update_users'                => false,
			'deactivate_old_memberships'  => false,
			'create_order'                => false,
			'partial'                     => false,
			'per_partial'                 => $this->per_partial,
			'site_id'                     => null,
		);
		
		$defaults = apply_filters( 'pmp_im_import_default_settings', $defaults );
		
		// Securely extract variables
		$settings = wp_parse_args( $args, $defaults );
		
		// Default new user notification target
		$msg_target = 'admin';
		
		// Cast variables to expected type
		$password_nag                = (bool) $settings['password_nag'];
		$new_user_notification       = (bool) $settings['new_user_notification'];
		$admin_new_user_notification = (bool) $settings['admin_new_user_notification'];
		$password_hashing_disabled   = (bool) $settings['password_hashing_disabled'];
		$update_users                = (bool) $settings['update_users'];
		$deactivate_old_memberships  = (bool) $settings['deactivate_old_memberships'];
		$suppress_pwdmsg             = (bool) $settings['suppress_pwdmsg'];
		$partial                     = (bool) $settings['partial'];
		$site_id                     = $settings['site_id'];
		$per_partial                 = apply_filters( 'pmp_im_import_records_per_scan', intval( $settings['per_partial'] ) );
		
		// User data fields list used to differentiate with user meta
		$userdata_fields = array(
			'ID',
			'user_login',
			'user_pass',
			'user_email',
			'user_url',
			'user_nicename',
			'display_name',
			'user_registered',
			'first_name',
			'last_name',
			'nickname',
			'description',
			'rich_editing',
			'comment_shortcuts',
			'admin_color',
			'use_ssl',
			'show_admin_bar_front',
			'show_admin_bar_admin',
			'role',
		);
		
		// Mac CR+LF fix
		ini_set( 'auto_detect_line_endings', true );
		
		$file              = basename( $filename );
		$this->file_object = new \SplFileObject( $filename, 'r' );
		
		// Use the expected delimiters, enclosures and escape characters
		$this->file_object->setCsvControl( PMP_IM_CSV_DELIMITER, PMP_IM_CSV_ENCLOSURE, PMP_IM_CSV_ESCAPE );
		$this->file_object->setFlags( \SplFileObject::READ_AHEAD | \SplFileObject::DROP_NEW_LINE | \SplFileObject::SKIP_EMPTY );
		
		// Loop through the file lines
		$first              = true;
		$line_number        = 0;
		$active_line_number = 1;
		$no_login           = false;
		
		if ( is_multisite() && ! empty( $site_id ) ) {
			switch_to_blog( $site_id );
		}
		
		while ( ( ! $this->file_object->eof() ) && ( true == $partial ? ( $line_number <= $per_partial ) : true ) ) {
			
			// Read a line from the file and remove the BOM character
			$line = $this->file_object->fgetcsv();
			
			// If the first line is empty, abort
			// If another line is empty, just skip it
			if ( empty( $line ) ) {
				
				if ( true === $first ) {
					$msg = __( "The expected header line in the import file is missing?!?", "pmpro-import-members-from-csv" );
					$this->add_error_msg( $msg, 'error' );
					$errors[] = new \WP_Error( 'pmp_im_header', $msg );
					break;
				} else {
					$line_number ++;
					continue;
				}
			}
			
			// If we are on the first line, the columns are the headers
			if ( true === $first ) {
				
				$headers = $this->process_header( $line );
				$first   = false;
				
				// Skip ahead ?
				if ( ! empty( $partial ) ) {
					
					// Get filename only
					$active_line_number = get_option( "pmpcsv_{$file}", null );
					
					// Skip to the proper line (during AJAX operations)
					if ( ! empty( $active_line_number ) ) {
						$this->file_object->seek( $active_line_number );
					}
				}
				
				// On to the next line in the file
				continue;
			} else {
				$active_line_number = $this->file_object->key();
			}
			
			// Suppress the email message to the user that their password was/may have been changed
			if ( true === $suppress_pwdmsg ) {
				add_filter( 'send_email_change_email', '__return_false', 99 );
				add_filter( 'send_password_change_email', '__return_false', 99 );
			}
			
			// Separate user data from meta
			$userdata = $usermeta = array();
			
			foreach ( $line as $ckey => $column ) {
				
				if ( ! isset( $headers[ $ckey ] ) ) {
					
					$msg = sprintf( __( "Cannot find header (column) %s!", "" ), $ckey );
					$this->add_error_msg( $msg, 'error' );
					$errors[] = new \WP_Error( 'pmp_im_header', $msg );
					
					if ( WP_DEBUG ) {
						error_log( $msg );
					}
					$line_number ++;
					continue;
				}
				
				$column_name = $headers[ $ckey ];
				$column      = trim( $column );
				
				if ( in_array( $column_name, $userdata_fields ) ) {
					$userdata[ $column_name ] = $column;
				} else {
					$usermeta[ $column_name ] = $column;
				}
			}
			
			if ( ! isset( $userdata['user_login'] ) && ! isset( $userdata['ID'] ) && ! isset( $userdata['user_email'] ) ) {
				
				$msg = __( "Missing ID, user_login or user_email information column", 'pmpro-import-members-from-csv' );
				$this->add_error_msg( $msg, 'error' );
				
				$errors[] = new \WP_Error( 'pmp_im_missing_data', $msg );
			}
			
			// A plugin may need to filter the data and meta
			$userdata = apply_filters( 'is_iu_import_userdata', $userdata, $usermeta, $settings );
			$userdata = apply_filters( 'pmp_im_import_userdata', $userdata, $usermeta, $settings );
			$usermeta = apply_filters( 'is_iu_import_usermeta', $usermeta, $userdata, $settings );
			$usermeta = apply_filters( 'pmp_im_import_usermeta', $usermeta, $userdata, $settings );
			
			// If no user data, bailout!
			if ( empty( $userdata ) ) {
				$msg = sprintf( __( "No user data found at row #%d", "pmpro-import-members-from-csv" ), ( $active_line_number + 1 ) );
				
				$warnings[] = new \WP_Error( 'pmp_im_nodata', $msg );
				
				if ( WP_DEBUG ) {
					error_log( $msg );
				}
				$line_number ++;
				continue;
			}
			
			// Something to be done before importing one user?
			do_action( 'is_iu_pre_user_import', $userdata, $usermeta );
			do_action( 'pmp_im_pre_member_import', $userdata, $usermeta );
			
			$user = $user_id = false;
			
			if ( isset( $userdata['ID'] ) ) {
				$user = get_user_by( 'ID', $userdata['ID'] );
			}
			
			if ( empty( $user ) && true == $update_users ) {
				
				if ( empty( $user ) && isset( $userdata['user_email'] ) ) {
					$user = get_user_by( 'email', $userdata['user_email'] );
				}
				
				if ( empty( $user ) && isset( $userdata['user_login'] ) ) {
					$user = get_user_by( 'login', $userdata['user_login'] );
				}
			}
			
			$update = false;
			
			if ( ! empty( $user ) ) {
				$userdata['ID'] = $user->ID;
				$update         = true;
			}
			
			// Generate a custom username (based on email address
			if ( ! in_array( 'user_login', $headers ) && empty( $user ) && ( ! empty( $userdata['user_email'] ) && is_email( $userdata['user_email'] ) ) ) {
				$msg                    = sprintf( __( "Created user login field for record at row %d", "pmpro-import-members-from-csv" ), ( $active_line_number + 1 ) );
				$warnings[]             = new \WP_Error( 'pmp_im_login', $msg );
				$userdata['user_login'] = preg_replace( '/-\.\_/', '', preg_replace( '/@.*/', '', $userdata['user_email'] ) );
			}
			
			if ( empty( $userdata['user_email'] ) || ( ! empty( $userdata['user_email'] ) && ! is_email( $userdata['user_email'] ) ) ) {
				
				$msg        = sprintf( __( "Invalid email in row %d (Not imported).", 'pmpro-import-members-from-csv' ), ( $active_line_number + 1 ) );
				$warnings[] = new \WP_Error( "pmp_im_email", $msg, isset( $userdata['user_email'] ) ? $userdata['user_email'] : null );
				$line_number ++;
				continue;
			}
			
			// BUG FIX: Not loading/updating record if user exists and the user identifiable data is the email address
			if ( ( empty( $userdata['user_login'] ) && empty( $userdata['user_email'] ) ) ||
			     ( ! empty( $userdata['user_login'] ) && false === $update && false !== get_user_by( 'login', $userdata['user_login'] ) ) ||
			     ( ! empty( $userdata['user_email'] ) && false === $update && false !== get_user_by( 'email', $userdata['user_email'] ) )
			) {
				$msg      = sprintf( __( "No user_login or user_email supplied/generated, or the user_login/user_email is already in use at row %d (Not imported!)", "pmpro-import-members-from-csv" ), $active_line_number + 1 );
				$errors[] = new \WP_Error( 'pmp_im_login', $msg );
				$line_number ++;
				continue;
			}
			
			// If creating a new user and no password was set, let auto-generate one!
			if ( false === $update && empty( $userdata['user_pass'] ) ) {
				$userdata['user_pass'] = wp_generate_password( 12, false );
			}
			
			// Insert, Update or insert without (re) hashing the password
			if ( true === $update && false === $password_hashing_disabled ) {
				$user_id = wp_update_user( $userdata );
			} else if ( false === $update && false === $password_hashing_disabled ) {
				$user_id = wp_insert_user( $userdata );
			} else if ( true === $password_hashing_disabled ) {
				$user_id = $this->insert_disabled_hashing_user( $userdata );
			} else {
				$errors[] = new \WP_Error( 'pmp_im_account', sprintf( __( 'No update/insert action taken for %s', 'pmpro-import-members-from-csv' ), $userdata['user_email'] ) );
				$line_number ++;
				continue;
			}
			
			$default_role = apply_filters( 'pmp_im_import_default_user_role', 'subscriber', $user_id, $site_id );
			
			// Is there an error?
			if ( is_wp_error( $user_id ) ) {
				$errors[ $line_number ] = $user_id;
			} else {
				
				// If no error, let's update the user meta too!
				if ( ! empty( $usermeta ) ) {
					foreach ( $usermeta as $metakey => $metavalue ) {
						$metavalue = maybe_unserialize( $metavalue );
						update_user_meta( $user_id, $metakey, $metavalue );
					}
				}
				
				// Set the password nag as needed
				if ( true === $password_nag ) {
					update_user_option( $user_id, 'default_password_nag', true, true );
				}
				
				// Adds the user to the specified blog ID if we're in a multisite configuration
				if ( is_multisite() && ! empty( $this->site_id ) ) {
					add_user_to_blog( $site_id, $user_id, $default_role );
				}
				
				// If we created a new user, send new user notification?
				if ( false === $update ) {
					
					// Only to the user?
					if ( true === $new_user_notification && false === $admin_new_user_notification ) {
						$msg_target = 'user';
					}
					
					// Only to the admin?
					if ( false === $new_user_notification && true === $admin_new_user_notification ) {
						$msg_target = 'admin';
					}
					
					// To the user _and_ the admin?
					if ( true === $new_user_notification && true === $admin_new_user_notification ) {
						$msg_target = 'both';
					}
					
					if ( true === $new_user_notification || true === $admin_new_user_notification ) {
						wp_new_user_notification( $user_id, null, $msg_target );
					}
				}
				
				if ( ! empty( $userdata['user_registered'] ) && true === $this->validate_date( $userdata['user_registered'] ) ) {
					
					// Update/set the user_registered value if the user is registered already.
					$update_registered = array(
						'ID'              => $user_id,
						'user_registered' => $userdata['user_registered'],
					);
					
					$status = wp_update_user( $update_registered );
					
					if ( is_wp_error( $status ) ) {
						$errors[ $line_number ] = $status;
						
						$should_be = $this->convert_time( $userdata['user_registered'] );
						$should_be = false === $should_be ? current_time( 'timestamp' ) : $should_be;
						
						$this->display_errors['user_registered'] = sprintf( __( 'The %2$suser_registered column%3$s contains an unrecognized date/time format. (Your format: \'%1$s\'. Expected format: \'%4$s\')', 'pmpro-import-members-from-csv' ), $userdata['user_registered'], '<strong>', '</strong>', date( 'Y-m-d h:i:s', $should_be ) );
					}
				}
				
				// Some plugins may need to do things after one user has been imported. Who know?
				do_action( 'is_iu_post_user_import', $user_id, $settings );
				do_action( 'pmp_im_post_member_import', $user_id, $settings );
				
				$user_ids[] = $user_id;
			}
			
			// Doing a partial import, save our location and then exit
			if ( ! empty( $partial ) && ! empty( $active_line_number ) ) {
				
				$active_line_number = ( $this->file_object->key() + 1 );
				
				update_option( "pmpcsv_{$file}", $active_line_number, 'no' );
			}
			
			$line_number ++;
		}
		
		// Close the file (done by the destructor for the SplFileObject() class)
		$this->file_object = null;
		ini_set( 'auto_detect_line_endings', true );
		
		// One more thing to do after all imports?
		// do_action( 'is_iu_post_users_import', $user_ids, $errors );
		do_action( 'pmp_im_post_members_import', $user_ids, $errors );
		
		// Let's log the errors
		$this->log_errors( array_merge( $errors, $warnings ) );
		
		// Return to the active (pre import) site
		if ( is_multisite() ) {
			switch_to_blog( $current_blog_id );
		}
		
		$member_error = (bool) get_option( 'e20r_import_errors', false );
		
		if ( true === $member_error ) {
			$this->add_error_msg( __( 'Data format error(s) detected during the import. Some records may not have been imported!', 'pmpro-import-members-from-csv' ), 'error' );
			delete_option( 'e20r_import_errors' );
		}
		
		// delete_option( "pmpcsv_{$file}" );
		
		return array(
			'user_ids' => $user_ids,
			'errors'   => $errors,
			'warnings' => $warnings,
		);
	}
	
	/**
	 * @param string[] $line
	 *
	 * @return string[]
	 */
	public function process_header( $line ) {
		
		$headers = $this->strip_BOM( $line );
		
		// Remove empty/blank headers
		foreach ( $headers as $hk => $hdr ) {
			if ( empty( $hdr ) ) {
				unset( $headers[ $hk ] );
			}
		}
		
		return $headers;
	}
	
	/**
	 * Strip UTF BOM character from line of text
	 *
	 * @param string[] $text_array
	 *
	 * @return string[]
	 *
	 * @since 2.9 - ENHANCEMENT: Strip away any BOM characters
	 */
	private function strip_BOM( $text_array ) {
		
		// Clear the old (possible) BOM 'infected' key
		$BOM           = pack( 'H*', 'EFBBBF' );
		$text_array[0] = preg_replace( "/^{$BOM}/", '', $text_array[0] );
		reset( $text_array );
		
		return $text_array;
	}
	
	/**
	 * Insert an user into the database.
	 * Copied from wp-include/user.php and commented wp_hash_password part
	 *
	 * @param mixed $userdata
	 *
	 * @return int|\WP_Error
	 *
	 * @since 2.0.1
	 *
	 **/
	private function insert_disabled_hashing_user( $userdata ) {
		
		global $wpdb;
		
		if ( is_a( $userdata, 'stdClass' ) ) {
			$userdata = get_object_vars( $userdata );
		} else if ( is_a( $userdata, 'WP_User' ) ) {
			$userdata = $userdata->to_array();
		}
		// Are we updating or creating?
		if ( ! empty( $userdata['ID'] ) ) {
			$ID            = (int) $userdata['ID'];
			$update        = true;
			$old_user_data = \WP_User::get_data_by( 'id', $ID );
			// hashed in wp_update_user(), plaintext if called directly
			// $user_pass = $userdata['user_pass'];
		} else {
			$update = false;
			// Hash the password
			// $user_pass = wp_hash_password( $userdata['user_pass'] );
		}
		$user_pass            = $userdata['user_pass'];
		$sanitized_user_login = sanitize_user( $userdata['user_login'], true );
		/**
		 * Filter a username after it has been sanitized.
		 *
		 * This filter is called before the user is created or updated.
		 *
		 * @since 2.0.1
		 *
		 * @param string $sanitized_user_login Username after it has been sanitized.
		 */
		$pre_user_login = apply_filters( 'pre_user_login', $sanitized_user_login );
		//Remove any non-printable chars from the login string to see if we have ended up with an empty username
		$user_login = trim( $pre_user_login );
		if ( empty( $user_login ) ) {
			return new \WP_Error( 'empty_user_login', __( 'Cannot create a user with an empty login name.', 'pmpro-import-members-from-csv' ) );
		}
		if ( false === $update && username_exists( $user_login ) ) {
			return new \WP_Error( 'existing_user_login', sprintf( __( 'Sorry, that username (%s) already exists!', 'pmpro-import-members-from-csv' ), $user_login ) );
		}
		if ( empty( $userdata['user_nicename'] ) ) {
			$user_nicename = sanitize_title( $user_login );
		} else {
			$user_nicename = $userdata['user_nicename'];
		}
		// Store values to save in user meta.
		$meta = array();
		/**
		 * Filter a user's nicename before the user is created or updated.
		 *
		 * @since 2.0.1
		 *
		 * @param string $user_nicename The user's nicename.
		 */
		$user_nicename = apply_filters( 'pre_user_nicename', $user_nicename );
		$raw_user_url  = empty( $userdata['user_url'] ) ? '' : $userdata['user_url'];
		/**
		 * Filter a user's URL before the user is created or updated.
		 *
		 * @since 2.0.1
		 *
		 * @param string $raw_user_url The user's URL.
		 */
		$user_url       = apply_filters( 'pre_user_url', $raw_user_url );
		$raw_user_email = empty( $userdata['user_email'] ) ? '' : $userdata['user_email'];
		/**
		 * Filter a user's email before the user is created or updated.
		 *
		 * @since 2.0.1
		 *
		 * @param string $raw_user_email The user's email.
		 */
		$user_email = apply_filters( 'pre_user_email', $raw_user_email );
		if ( false === $update && ! defined( 'WP_IMPORTING' ) && email_exists( $user_email ) ) {
			return new \WP_Error( 'existing_user_email', sprintf( __( 'Sorry, that email address (%s) is already used!', 'pmpro-import-members-from-csv' ), $user_email ) );
		}
		$nickname = empty( $userdata['nickname'] ) ? $user_login : $userdata['nickname'];
		/**
		 * Filter a user's nickname before the user is created or updated.
		 *
		 * @since 2.0.1
		 *
		 * @param string $nickname The user's nickname.
		 */
		$meta['nickname'] = apply_filters( 'pre_user_nickname', $nickname );
		$first_name       = empty( $userdata['first_name'] ) ? '' : $userdata['first_name'];
		/**
		 * Filter a user's first name before the user is created or updated.
		 *
		 * @since 2.0.1
		 *
		 * @param string $first_name The user's first name.
		 */
		$meta['first_name'] = apply_filters( 'pre_user_first_name', $first_name );
		$last_name          = empty( $userdata['last_name'] ) ? '' : $userdata['last_name'];
		/**
		 * Filter a user's last name before the user is created or updated.
		 *
		 * @since 2.0.1
		 *
		 * @param string $last_name The user's last name.
		 */
		$meta['last_name'] = apply_filters( 'pre_user_last_name', $last_name );
		if ( empty( $userdata['display_name'] ) ) {
			if ( true === $update ) {
				$display_name = $user_login;
			} else if ( $meta['first_name'] && $meta['last_name'] ) {
				/* translators: 1: first name, 2: last name */
				$display_name = sprintf( _x( '%1$s %2$s', 'Display name based on first name and last name', 'pmpro-import-members-from-csv' ), $meta['first_name'], $meta['last_name'] );
			} else if ( $meta['first_name'] ) {
				$display_name = $meta['first_name'];
			} else if ( $meta['last_name'] ) {
				$display_name = $meta['last_name'];
			} else {
				$display_name = $user_login;
			}
		} else {
			$display_name = $userdata['display_name'];
		}
		/**
		 * Filter a user's display name before the user is created or updated.
		 *
		 * @since 2.0.1
		 *
		 * @param string $display_name The user's display name.
		 */
		$display_name = apply_filters( 'pre_user_display_name', $display_name );
		$description  = empty( $userdata['description'] ) ? '' : $userdata['description'];
		/**
		 * Filter a user's description before the user is created or updated.
		 *
		 * @since 2.0.1
		 *
		 * @param string $description The user's description.
		 */
		$meta['description']          = apply_filters( 'pre_user_description', $description );
		$meta['rich_editing']         = empty( $userdata['rich_editing'] ) ? 'true' : $userdata['rich_editing'];
		$meta['comment_shortcuts']    = empty( $userdata['comment_shortcuts'] ) ? 'false' : $userdata['comment_shortcuts'];
		$admin_color                  = empty( $userdata['admin_color'] ) ? 'fresh' : $userdata['admin_color'];
		$meta['admin_color']          = preg_replace( '|[^a-z0-9 _.\-@]|i', '', $admin_color );
		$meta['use_ssl']              = empty( $userdata['use_ssl'] ) ? 0 : $userdata['use_ssl'];
		$user_registered              = empty( $userdata['user_registered'] ) ? gmdate( 'Y-m-d H:i:s' ) : $userdata['user_registered'];
		$meta['show_admin_bar_front'] = empty( $userdata['show_admin_bar_front'] ) ? 'true' : $userdata['show_admin_bar_front'];
		$user_nicename_check          = $wpdb->get_var(
			$wpdb->prepare(
				"SELECT ID FROM {$wpdb->users} WHERE user_nicename = %s AND user_login != %s LIMIT 1",
				$user_nicename,
				$user_login
			)
		);
		
		if ( ! empty( $user_nicename_check ) ) {
			
			$suffix = 2;
			
			while ( $user_nicename_check ) {
				
				$alt_user_nicename   = $user_nicename . "-$suffix";
				$user_nicename_check = $wpdb->get_var(
					$wpdb->prepare(
						"SELECT ID FROM {$wpdb->users} WHERE user_nicename = %s AND user_login != %s LIMIT 1",
						$alt_user_nicename,
						$user_login
					)
				);
				$suffix ++;
			}
			
			$user_nicename = $alt_user_nicename;
		}
		
		$compacted = compact( 'user_pass', 'user_email', 'user_url', 'user_nicename', 'display_name', 'user_registered' );
		
		$data = wp_unslash( $compacted );
		
		if ( true === $update ) {
			$wpdb->update( $wpdb->users, $data, compact( 'ID' ) );
			$user_id = (int) $ID;
		} else {
			$wpdb->insert( $wpdb->users, $data + compact( 'user_login' ) );
			$user_id = (int) $wpdb->insert_id;
		}
		$user = new \WP_User( $user_id );
		// Update user meta.
		foreach ( $meta as $key => $value ) {
			update_user_meta( $user_id, $key, $value );
		}
		foreach ( wp_get_user_contact_methods( $user ) as $key => $value ) {
			if ( isset( $userdata[ $key ] ) ) {
				update_user_meta( $user_id, $key, $userdata[ $key ] );
			}
		}
		if ( isset( $userdata['role'] ) ) {
			$user->set_role( $userdata['role'] );
		} else if ( false === $update ) {
			$user->set_role( get_option( 'default_role' ) );
		}
		wp_cache_delete( $user_id, 'users' );
		wp_cache_delete( $user_login, 'userlogins' );
		if ( true === $update ) {
			/**
			 * Fires immediately after an existing user is updated.
			 *
			 * @since 2.0.1
			 *
			 * @param int    $user_id       User ID.
			 * @param object $old_user_data Object containing user's data prior to update.
			 */
			do_action( 'profile_update', $user_id, $old_user_data );
		} else {
			/**
			 * Fires immediately after a new user is registered.
			 *
			 * @since 2.0.1
			 *
			 * @param int $user_id User ID.
			 */
			do_action( 'user_register', $user_id );
		}
		
		return $user_id;
	}
	
	/**
	 * Test the date supplied for MySQL compliance
	 *
	 * @param string $date
	 * @param string $format
	 *
	 * @return bool
	 *
	 * @credit Stack Overflow: User @glavić - https://stackoverflow.com/a/12323025
	 */
	private function validate_date( $date, $format = 'Y-m-d H:i:s' ) {
		
		$check_date = \DateTime::createFromFormat( $format, $date );
		$retval     = ( false !== $check_date ) && ( $check_date->format( $format ) == $date );
		
		return $retval;
	}
	
	/**
	 * Attempt to locate a valid time for the specified time string
	 *
	 * @param string $time_string
	 *
	 * @return false|int
	 */
	private function convert_time( $time_string ) {
		
		$timestamp = strtotime( $time_string, current_time( 'timestamp' ) );
		
		if ( false === $timestamp ) {
			$timestamp = strtotime(
				preg_replace( '/\//', '-', $time_string ),
				current_time( 'timestamp' )
			);
		}
		
		return $timestamp;
	}
	
	/**
	 * Log errors to a file
	 *
	 * @param array $errors
	 *
	 * @since 1.0
	 **/
	private function log_errors( $errors ) {
		
		if ( empty( $errors ) ) {
			return;
		}
		
		$this->add_error_msg(
			sprintf(
				__( 'Errors or Warnings found: Please inspect the import %1$serror log%2$s', 'pmpro-import-members-from-csv' ),
				sprintf(
					'<a href="%1$s" title="%2$s" target="_blank">',
					esc_url_raw( $this->logfile_url ),
					__( "Link to import error log", "pmpro-import-members-from-csv" )
				),
				'</a>'
			),
			'warning'
		);
		
		if ( false !== ( $log = fopen( $this->logfile_path, 'a' ) ) ) {
			
			/**
			 * @param \WP_Error $error
			 */
			foreach ( $errors as $key => $error ) {
				$line = $key + 1;
				
				// Handle weird/unexpected formats for error message(s)
				if ( is_wp_error( $error ) ) {
					$message = $error->get_error_message();
				} else if ( is_string( $error ) ) {
					$message = $error;
				}
				
				@fwrite( $log, sprintf(
					               __( '[Line %1$s] %2$s', 'pmpro-import-members-from-csv' ), $line, $message ) . "\n" );
			}
			
			fclose( $log );
		} else {
			$this->add_error_msg( sprintf( __( "Unable to write to error log at %s", "pmpro-import-members-from-csv" ), $this->logfile_path ), 'error' );
		}
	}
	
	/**
	 * Remove handlers for the "Import Users from CSV Integration" add-on
	 */
	public function removeIUCSVSupport() {
		
		if ( has_action( 'is_iu_pre_user_import', 'pmproiufcsv_is_iu_pre_user_import' ) ) {
			
			remove_action( 'is_iu_pre_user_import', 'pmproiufcsv_is_iu_pre_user_import', 10, 2 );
			remove_action( 'is_iu_post_user_import', 'pmproiufcsv_is_iu_post_user_import' );
			remove_filter( 'is_iu_import_usermeta', 'pmproiufcsv_is_iu_import_usermeta', 10, 2 );
		}
		
	}
	
	/**
	 * Remove the import source file upon successful/complete import
	 */
	public function wp_ajax_cleanup_csv() {
		
		if ( WP_DEBUG ) {
			error_log( "Import is complete: " . print_r( $_REQUEST, true ) );
		}
		
		check_admin_referer( 'pmp-im-import-members', 'pmp-im-import-members-wpnonce' );
		
		if ( WP_DEBUG ) {
			error_log( "Nonce is good" );
		}
		
		$upload_dir = wp_upload_dir();
		$file_name  = isset( $_REQUEST['filename'] ) ? sanitize_file_name( $_REQUEST['filename'] ) : null;
		$file_name  = trailingslashit( $upload_dir['basedir'] ) . "e20r_imports/{$file_name}";
		$file       = basename( $file_name );
		
		if ( ! is_null( $file_name ) && file_exists( $file_name ) ) {
			unlink( $file_name );
			delete_option( "pmpcsv_{$file}" );
			
			wp_send_json_success();
			exit();
		}
		
		wp_send_json_error();
		exit();
	}
	
	/**
	 * AJAX service that does the heavy loading to import a CSV file
	 *
	 * @since 2.0
	 */
	public function wp_ajax_import_members_from_csv() {
		
		//get settings
		$this->load_settings();
		
		// Error message to return
		if ( empty( $this->filename ) ) {
			wp_send_json_error( array(
				'status'  => - 1,
				'message' => __( "No import file provided!", "pmpro-import-members-from-csv" ),
			) );
			exit();
		}
		
		//figure out upload dir
		$upload_dir = wp_upload_dir();
		$import_dir = $upload_dir['basedir'] . "/e20r_imports/";
		
		//make sure file exists
		if ( ! file_exists( "{$import_dir}{$this->filename}" ) ) {
			wp_send_json_error(
				array(
					'status'  => - 1,
					'message' => sprintf(
						__( "File (%s) not found in %s\nIs the directory writable by the web server software?", 'pmpro-import-members-from-csv' ),
						$this->filename,
						$import_dir
					),
				)
			);
			exit();
		}
		
		//import next few lines of file
		$args = array(
			'partial'                     => true,
			'filename'                    => $this->filename,
			'password_nag'                => $this->password_nag,
			'password_hashing_disabled'   => $this->password_hashing_disabled,
			'update_users'                => $this->update_users,
			'new_user_notification'       => $this->new_user_notification,
			'new_member_notification'     => $this->new_member_notification,
			'admin_new_user_notification' => $this->admin_new_user_notification,
			'suppress_pwdmsg'             => $this->suppress_pwdmsg,
			'send_welcome_email'          => $this->send_welcome_email,
			'deactivate_old_memberships'  => $this->deactivate_old_memberships,
			'background_import'           => $this->background_import,
			'per_partial'                 => $this->per_partial,
			'site_id'                     => $this->site_id,
			'create_order'                => $this->create_order,
		);
		
		$args = apply_filters( 'pmp_im_import_arguments', $args );
		
		if ( WP_DEBUG ) {
			error_log( "Path to import file: {$import_dir}{$this->filename}" );
		}
		
		$results = $this->import_csv( "{$import_dir}{$this->filename}", $args );
		
		$error_log_msg = null;
		
		if ( file_exists( $this->logfile_path ) ) {
			$error_log_msg = sprintf(
				__( ', please %1$scheck the error log%2$s', 'pmpro-import-members-from-csv' ),
				sprintf( '<a href="%s">', esc_url_raw( $this->logfile_url ) ),
				'</a>'
			);
		}
		
		if ( isset( $_REQUEST['import'] ) ) {
			$error_log_msg = '';
			
			if ( file_exists( $this->logfile_path ) ) {
				$error_log_msg = sprintf(
					__( ', please %1$scheck the error log%2$s', 'pmpro-import-members-from-csv' ),
					sprintf( '<a href="%s">', esc_url_raw( $this->logfile_url ) ),
					'</a>'
				);
			}
			
			switch ( $_REQUEST['import'] ) {
				case 'file':
					$status = sprintf( '<div class="error"><p><strong>%s</strong></p></div>', __( 'Error during file upload.', 'pmpro-import-members-from-csv' ) );
					break;
				case 'data':
					$status = sprintf( '<div class="error"><p><strong>%s</strong></p></div>', __( 'Cannot extract data from uploaded file or no file was uploaded.', 'pmpro-import-members-from-csv' ) );
					break;
				case 'success':
					$status = sprintf( '<div class="updated"><p><strong>%s</strong></p></div>', __( 'Member import was successful.', 'pmpro-import-members-from-csv' ) );
					break;
				default:
					$status = null;
			}
		}
		
		$buffered_text = ob_get_clean();
		
		// No users imported (or done)
		if ( empty( $results['user_ids'] ) ) {
			
			//Clear the file
			unlink( "{$import_dir}{$this->filename}" );
			
			//Clear position
			$file = basename( $this->filename );
			delete_option( "pmpcsv_{$file}" );
			
			wp_send_json_success( array(
				'status'         => true,
				'message'        => $status,
				'display_errors' => ( ! empty( $this->display_errors ) ? $this->display_errors : null ),
			) );
			exit();
			
		} else if ( ! empty( $results['errors'] ) ) {
			
			/**
			 * @param string[] $msgs
			 */
			$msgs = array();
			
			/**
			 * @param \WP_Error $error
			 */
			foreach ( $results['errors'] as $error ) {
				$msgs[] = $error->get_error_message();
			}
			
			wp_send_json_error( array(
				'status'         => false,
				'message'        => sprintf( __( "Error during import (# of errors: %d):\n%s", 'pmpro-import-members-from-csv' ), count( $msgs ), implode( "\n", $msgs ) ),
				'display_errors' => ( ! empty( $this->display_errors ) ? $this->display_errors : null ),
			) );
			exit();
		} else {
			
			/**
			 * @param string[] $msgs
			 */
			$msgs = array();
			
			if ( ! empty( $results['warnings'] ) ) {
				
				/**
				 * @param \WP_Error $error
				 */
				foreach ( $results['warnings'] as $error ) {
					$msgs[] = $error->get_error_message();
				}
				
				if ( WP_DEBUG ) {
					error_log( "Warnings: " . print_r( $msgs, true ) );
				}
			}
			
			$status_msg = sprintf( __( "Imported %s", "pmpro-import-members-from-csv" ), str_pad( '', count( $results['user_ids'] ), '.' ) ) . "\n";
			
			if ( ! empty( $msgs ) ) {
				$status_msg .= implode( "\n", $msgs ) . "\n";
			}
			
			wp_send_json_success( array(
				'status'         => true,
				'message'        => $status_msg,
				'display_errors' => ( ! empty( $this->display_errors ) ? $this->display_errors : null ),
			) );
			exit();
		}
	}
	
	/**
	 * Delete all import_ meta fields before an import in case the user has been imported in the past.
	 *
	 * @param array $user_data
	 * @param array $user_meta
	 */
	public function pre_member_import( $user_data, $user_meta ) {
		
		// Init variables
		$user   = false;
		$target = null;
		
		//Get user by ID
		if ( isset( $user_data['ID'] ) ) {
			$user = get_user_by( 'ID', $user_data['ID'] );
		}
		
		// That didn't work, now try by login value or email
		if ( empty( $user->ID ) ) {
			
			if ( isset( $user_data['user_login'] ) ) {
				$target = 'login';
				
			} else if ( isset( $user_data['user_email'] ) ) {
				$target = 'email';
			}
			
			if ( ! empty( $target ) ) {
				$user = get_user_by( $target, $user_data["user_{$target}"] );
			} else {
				return; // Exit quietly
			}
		}
		
		// Clean up if we found a user (delete the imported_ usermeta)
		if ( ! empty( $user->ID ) ) {
			
			foreach ( $this->pmpro_fields as $field_name => $value ) {
				delete_user_meta( $user->ID, "imported_{$field_name}" );
			}
		}
	}
	
	/**
	 * Change some of the imported columns to add "imported_" to the front so we don't confuse the data later.
	 *
	 * @param array $user_meta
	 * @param array $user_data
	 *
	 * @return array
	 */
	public function import_usermeta( $user_meta, $user_data ) {
		
		foreach ( $user_meta as $key => $value ) {
			
			if ( in_array( $key, array_keys( $this->pmpro_fields ) ) ) {
				$key = "imported_{$key}";
			}
			
			$user_meta[ $key ] = $value;
		}
		
		return $user_meta;
	}
	
	/**
	 * Clean up by removing unused imported_* fields
	 *
	 * @param int   $user_id
	 * @param array $settings
	 *
	 * @since 2.9 ENHANCEMENT: Clean up unneeded user metadata
	 */
	public function cleanup( $user_id, $settings ) {
		
		foreach ( $this->pmpro_fields as $field_name => $value ) {
			delete_user_meta( $user_id, "imported_{$field_name}" );
		}
	}
	
	/**
	 * After the new user was created, import PMPro membership metadata
	 *
	 * @param int   $user_id
	 * @param array $settings
	 *
	 * @since 2.9 BUG FIX: A little too silent when the imported file is mis-configured
	 * @since 2.9 BUG FIX: MS Excel causing trouble w/first column import values (Improved UTF BOM handling)
	 * @since 2.9 ENHANCEMENT: Improved error logging/handling for typical import file errors
	 * @since 2.9 ENHANCEMENT: Send Welcome Email template when adding user to membership level ("Welcome imported
	 *        user" email located in emails/ directory)
	 * @since 2.9 BUG FIX: Typo in warning/error messages
	 * @since 2.22 ENHANCEMENT: Move data checks to own method (validate_membership_data())
	 *
	 */
	public function import_membership_info( $user_id, $settings = array() ) {
		
		global $wpdb;
		$errors = array();
		
		$has_error              = false;
		$membership_in_the_past = false;
		
		// Define table names
		$pmpro_member_table  = "{$wpdb->prefix}pmpro_memberships_users";
		$pmpro_dc_table      = "{$wpdb->prefix}pmpro_discount_codes";
		$pmpro_dc_uses_table = "{$wpdb->prefix}pmpro_discount_codes_uses";
		
		if ( is_multisite() ) {
			$current_blog_id = get_current_blog_id();
		}
		
		wp_cache_delete( $user_id, 'users' );
		$user = get_userdata( $user_id );
		
		if ( empty( $user ) ) {
			$errors[]  = new \WP_Error( 'pmp_im_member', sprintf( __( "Unable to locate user with expected user ID of %d (Skipped!)", "pmpro-import-members-from-csv" ), $user_id ) );
			$has_error = true;
		}
		
		// Generate PMPro specific member value(s)
		foreach ( $this->pmpro_fields as $var_name => $field_value ) {
			$this->pmpro_fields[ $var_name ] = isset( $user->{"imported_{$var_name}"} ) ? $user->{"imported_{$var_name}"} : $this->pmpro_fields[ $var_name ];
		}
		
		if ( WP_DEBUG ) {
			error_log( "Adding membership info for: {$user_id}" );
		}
		
		// Set site ID and custom table names for the multi site configs
		if ( is_multisite() ) {
			switch_to_blog( $this->site_id );
			
			$pmpro_member_table  = "{$wpdb->base_prefix}pmpro_memberships_users";
			$pmpro_dc_table      = "{$wpdb->base_prefix}pmpro_discount_codes";
			$pmpro_dc_uses_table = "{$wpdb->base_prefix}pmpro_discount_codes_uses";
			
		}
		
		$has_error = $this->validate_membership_data( $user_id );
		
		/**
		 * @since v2.41 - BUG FIX: Didn't warn when membership_status wasn't included but user
		 * wanted to send Welcome email
		 */
		if ( empty( $this->welcome_mail_warning ) &&
		     true == $this->send_welcome_email &&
		     ( ! isset( $this->pmpro_fields['membership_status'] ) ||
		       ( isset( $this->pmpro_fields['membership_status'] ) && 'active' != trim( $this->pmpro_fields['membership_status'] ) )
		     )
		) {
			$this->welcome_mail_warning = __(
			        'Cannot send Welcome email to members who did not get imported as \'active\' members!',
                    'pmpro-import-members-from-csv'
            );
			
			if ( WP_DEBUG ) {
			    error_log( $this->welcome_mail_warning);
            }
			$this->add_error_msg( $this->welcome_mail_warning, 'warning' );
		}
		
		// Proceed to import member data if the data is correct
		if ( false === $has_error ) {
			
			//Look up the discount code when included
			if ( ! empty( $this->pmpro_fields['membership_discount_code'] ) && empty( $this->pmpro_fields['membership_code_id'] ) ) {
				
				$this->pmpro_fields['membership_code_id'] = $wpdb->get_var(
					$wpdb->prepare(
						"SELECT dc.id
                              FROM {$pmpro_dc_table} AS dc
                              WHERE dc.code = %s
                              LIMIT 1",
						$this->pmpro_fields['membership_discount_code']
					)
				);
			}
			
			//Change membership level
			if ( ( is_numeric( $this->pmpro_fields['membership_id'] ) && 0 <= $this->pmpro_fields['membership_id'] ) || ! empty( $this->pmpro_fields['membership_id'] ) ) {
				
				// Cancel previously existing (active) memberships (Should support MMPU add-on)
				// without triggering cancellation emails, etc
				if ( true === $this->deactivate_old_memberships ) {
					
					// Update all currently active memberships with the specified ID for the specified user
					if ( false === ( $updated = $wpdb->update( $pmpro_member_table, array( 'status' => 'admin_cancelled' ), array(
							'user_id'       => $user_id,
							'membership_id' => $this->pmpro_fields['membership_id'],
							'status'        => 'active',
						) ) ) ) {
						
						$errors[] = new \WP_Error( 'pmp_im_member', sprintf(
							__( 'Unable to cancel old membership level (ID: %d) for user (ID: %d)', 'pmpro-import-members-from-csv' ),
							$this->pmpro_fields['membership_id'],
							$user_id
						) );
					}
				}
				
				$custom_level = array(
					'user_id'         => $user_id,
					'membership_id'   => $this->pmpro_fields['membership_id'],
					'code_id'         => ! empty( $this->pmpro_fields['membership_code_id'] ) ? $this->pmpro_fields['membership_code_id'] : 0,
					'initial_payment' => ! empty( $this->pmpro_fields['membership_initial_payment'] ) ? $this->pmpro_fields['membership_initial_payment'] : '',
					'billing_amount'  => ! empty( $this->pmpro_fields['membership_billing_amount'] ) ? $this->pmpro_fields['membership_billing_amount'] : '',
					'cycle_number'    => ! empty( $this->pmpro_fields['membership_cycle_number'] ) ? $this->pmpro_fields['membership_cycle_number'] : '',
					'cycle_period'    => ! empty( $this->pmpro_fields['membership_cycle_period'] ) ? $this->pmpro_fields['membership_cycle_period'] : 'Month',
					'billing_limit'   => ! empty( $this->pmpro_fields['membership_billing_limit'] ) ? $this->pmpro_fields['membership_billing_limit'] : '',
					'trial_amount'    => ! empty( $this->pmpro_fields['membership_trial_amount'] ) ? $this->pmpro_fields['membership_trial_amount'] : '',
					'trial_limit'     => ! empty( $this->pmpro_fields['membership_trial_limit'] ) ? $this->pmpro_fields['membership_trial_limit'] : '',
					'startdate'       => ! empty( $this->pmpro_fields['membership_startdate'] ) ? $this->pmpro_fields['membership_startdate'] : null,
				);
				
				// Add the enddate
				$custom_level['enddate'] = ! empty( $this->pmpro_fields['membership_enddate'] ) ? $this->pmpro_fields['membership_enddate'] : null;
				
				// Set the status of the membership
				if ( ! empty( $this->pmpro_fields['membership_status'] ) ) {
					$custom_level['status'] = $this->pmpro_fields['membership_status'];
				}
				
				// $updated_level = $this->update_membership_level( $custom_level, $user_id );
				
				/**
				 * @since v2.50 - BUG FIX: Don't deactivate old levels (here)
				 */
				add_filter( 'pmpro_deactivate_old_levels', '__return_false', 999 );
				add_filter( 'pmpro_cancel_previous_subscriptions', '__return_false', 999 );
				
				// Update the level
				$updated_level = pmpro_changeMembershipLevel( $custom_level, $user_id, 'cancelled' );
				
				remove_filter( 'pmpro_deactivate_old_levels', '__return_false', 999 );
				remove_filter( 'pmpro_cancel_previous_subscriptions', '__return_false', 999 );
				
				if ( false === $updated_level ) {
					
					$errors[] = new \WP_Error(
						'pmp_im_member',
						sprintf( __(
							'Unable to configure membership level (ID: %d ) for user (user id: %d)',
							'pmpro-import-members-from-csv'
						),
							$this->pmpro_fields['membership_id'],
							$user_id
						)
					);
					
				}
				
				if ( ! empty( $custom_level['status'] ) && $custom_level['status'] != 'inactive' ) {
					// Get the most recently added column
					$record_id = $wpdb->get_var(
						$wpdb->prepare(
							"SELECT mt.id
                                      FROM {$pmpro_member_table} AS mt
                                      WHERE mt.user_id = %d AND mt.membership_id = %d AND mt.status = %s
                                      ORDER BY mt.id DESC LIMIT 1",
							$user_id,
							$custom_level['membership_id'],
							$custom_level['status']
						)
					);
				} else {
					$record_id = $wpdb->get_var(
						$wpdb->prepare(
							"SELECT mt.id
                                      FROM {$pmpro_member_table} AS mt
                                      WHERE mt.user_id = %d AND mt.membership_id = %d
                                      ORDER BY mt.id DESC LIMIT 1",
							$user_id,
							$custom_level['membership_id']
						)
					);
				}
				
				// If membership ended in the past, make it inactive for now
				if ( ( "inactive" == strtolower( $this->pmpro_fields['membership_status'] ) && ! empty( $record_id ) ) ||
				     ( ! empty( $this->pmpro_fields['membership_enddate'] ) &&
				       strtoupper( $this->pmpro_fields['membership_enddate'] ) != "NULL" &&
				       strtotime( $this->pmpro_fields['membership_enddate'], current_time( 'timestamp' ) ) < current_time( 'timestamp' )
				     )
				) {
					
					/**
					 * @since 2.11 - BUG FIX: Didn't handle 'inactive' status with no membership_enddate supplied
					 */
					if ( empty( $this->pmpro_fields['membership_enddate'] ) ) {
						$this->pmpro_fields['membership_enddate'] = date( 'Y-m-d h:i:s', current_time( 'timestamp' ) );
					}
					
					if ( false !== $wpdb->update(
							$pmpro_member_table,
							array( 'status' => 'inactive', 'enddate' => $this->pmpro_fields['membership_enddate'] ),
							array(
								'id'            => $record_id,
								'user_id'       => $user_id,
								'membership_id' => $this->pmpro_fields['membership_id'],
							),
							array( '%s', '%s' ),
							array( '%d', '%d', '%d' )
						)
					) {
						$membership_in_the_past = true;
					} else {
						$errors[] = new \WP_Error( 'pmp_im_member', sprintf( __( 'Unable to set \'inactive\' membership status/date for user (ID: %d) with membership level ID %d', 'pmpro-import-members-from-csv' ), $user_id, $this->pmpro_fields['membership_id'] ) );
					}
				}
				
				if ( 'active' == strtolower( $this->pmpro_fields['membership_status'] ) &&
				     ( empty( $this->pmpro_fields['membership_enddate'] ) ||
				       'NULL' == strtoupper( $this->pmpro_fields['membership_enddate'] ) ||
				       strtotime( $this->pmpro_fields['membership_enddate'], current_time( 'timestamp' ) ) >= current_time( 'timestamp' ) )
				) {
					
					if ( false === $wpdb->update(
							$pmpro_member_table,
							array( 'status' => 'active', 'enddate' => $this->pmpro_fields['membership_enddate'] ),
							array(
								'id'            => $record_id,
								'user_id'       => $user_id,
								'membership_id' => $this->pmpro_fields['membership_id'],
							),
							array( '%s', '%s' ),
							array( '%d', '%d', '%d' )
						) ) {
						$errors[] = new \WP_Error( 'pmp_im_member', sprintf( __( 'Unable to activate membership for user (ID: %d) with membership level ID %d', 'pmpro-import-members-from-csv' ), $user_id, $this->pmpro_fields['membership_id'] ) );
					}
				}
			}
			
			// Add a PMPro order record so integration with gateway doesn't cause surprises
			if ( true === $this->create_order || (
					! empty( $this->pmpro_fields['membership_subscription_transaction_id'] ) && ! empty( $this->pmpro_fields['membership_gateway'] )
				)
			) {
				
				if ( WP_DEBUG ) {
					error_log( "Adding PMPro order for {$this->pmpro_fields['membership_id']}/{$user_id}" );
				}
				
				$default_gateway     = function_exists( 'pmpro_getGateway' ) ? pmpro_getGateway() : $this->pmpro_fields['membership_gateway'];
				$default_environment = function_exists( 'pmpro_getOption' ) ? pmpro_getOption( "gateway_environment" ) : $this->pmpro_fields['membership_gateway_environment'];
				
				/**
				 * Load order table columns for processing
				 */
				$order_fields = $this->get_table_info( 'pmpro_membership_orders' );
				
				if ( empty( $this->pmpro_fields['membership_initial_payment'] ) && empty( $this->pmpro_fields['membership_billing_amount'] ) && ! empty( $this->pmpro_fields['membership_id'] ) ) {
					
					$default_level = pmpro_getLevel( $this->pmpro_fields['membership_id'] );
					
					if ( ! empty( $default_level ) ) {
						$this->pmpro_fields['membership_initial_payment'] = $default_level->initial_payment;
						$this->pmpro_fields['membership_billing_amount']  = $default_level->billing_amount;
						$this->pmpro_fields['membership_billing_limit']   = $default_level->billing_limit;
						$this->pmpro_fields['membership_cycle_number']    = $default_level->cycle_number;
						$this->pmpro_fields['membership_cycle_period']    = $default_level->cycle_period;
					}
				}
				
				$order                 = new \MemberOrder();
				$order->user_id        = $user_id;
				$order->membership_id  = $this->pmpro_fields['membership_id'];
				$order->InitialPayment = ! empty( $this->pmpro_fields['membership_initial_payment'] ) ? $this->pmpro_fields['membership_initial_payment'] : null;
				
				/**
				 * Dynamically provide data for all configured Order fields...
				 */
				foreach ( $order_fields as $full_field_name => $default_value ) {
					
					$process_billing_info = false;
					
					if ( 'membership_id' !== strtolower( $full_field_name ) && 'user_id' !== strtolower( $full_field_name ) ) {
						$field_name = preg_replace( '/membership_/', '', strtolower( $full_field_name ) );
					} else {
						$field_name = strtolower( $full_field_name );
					}
					
					/**
					 * Add billing info (if/when we can)
					 */
					if ( 1 === preg_match( "/billing_(.*)/", $field_name, $matches ) ) {
						
						if ( ! isset( $order->billing ) ) {
							$order->billing = new \stdClass();
						}
						
						if ( ! isset( $order->billing->{$matches[1]} ) ) {
							
							$meta_key = $this->map_billing_field_to_meta( $field_name );
							
							if ( ! empty( $meta_key ) ) {
								$order->billing->{$matches[1]} = get_user_meta( $user_id, $meta_key, true );
								$process_billing_info          = true;
							}
						}
					}
					
					if ( false === $process_billing_info && ( ! isset( $order->{$field_name} ) || ( isset( $order->{$field_name} ) && empty( $order->{$field_name} ) ) ) ) {
						
						// Process payment (amount)
						if ( 'total' === $field_name && ! empty( $this->pmpro_fields['membership_initial_payment'] ) ) {
							
							$order->total = $this->pmpro_fields['membership_initial_payment'];
							
						} else if ( 'total' === $field_name && (
								empty( $this->pmpro_fields['membership_initial_payment'] ) &&
								! empty( $this->pmpro_fields['membership_billing_amount'] )
							) ) {
							
							$order->total = $this->pmpro_fields['membership_billing_amount'];
							
						} else if ( 'total' !== $field_name ) {
							
							if ( 'status' === $field_name ) {
								$order->{$field_name} = ( $this->pmpro_fields[ $full_field_name ] = 'active' ) ? 'success' : 'cancelled';
							} else {
								$order->{$field_name} = ! empty( $this->pmpro_fields[ $full_field_name ] ) ? $this->pmpro_fields[ $full_field_name ] : null;
							}
						} else {
							if ( WP_DEBUG ) {
								error_log( "Warning: {$field_name} will not be processed!!" );
							}
						}
					}
				}
				
				if ( strtolower( $default_gateway ) !== strtolower( $this->pmpro_fields['membership_gateway'] ) ) {
					$order->setGateway( $this->pmpro_fields['membership_gateway'] );
				}
				
				if ( strtolower( $default_environment ) !== strtolower( $this->pmpro_fields['membership_gateway_environment'] ) ) {
					$order->gateway_environment = strtolower( $this->pmpro_fields['membership_gateway_environment'] );
				}
				
				if ( true === $membership_in_the_past ) {
					$order->status = "cancelled";
				}
				
				if ( ! empty( $order->billing ) && WP_DEBUG ) {
					error_log( "Billing info: " . print_r( $order->billing, true ) );
				}
				
				/**
				 * Add MemberOrder billing info if possible
				 */
				if ( ! empty( $order->billing ) ) {
					$order->billing->name = "{$user->first_name} {$user->last_name}";
				}
				
				if ( false === $order->saveOrder() ) {
					$msg      = sprintf( __( 'Unable to save order object for user (ID: %d).', 'pmpro-import-members-from-csv' ), $user_id );
					$errors[] = new \WP_Error( 'pmp_im_member', $msg );
					
					if ( WP_DEBUG ) {
						error_log( $msg );
					}
				}
				
				//update timestamp of order?
				if ( ! empty( $this->pmpro_fields['membership_timestamp'] ) ) {
					
					if ( true === $this->validate_date( $this->pmpro_fields['membership_timestamp'], 'Y-m-d H:i:s' ) ) {
						$timestamp = strtotime( $this->pmpro_fields['membership_timestamp'], current_time( 'timestamp' ) );
					} else {
						$timestamp = is_numeric( $this->pmpro_fields['membership_timestamp'] ) ? $this->pmpro_fields['membership_timestamp'] : null;
						
						if ( is_null( $timestamp ) ) {
							$errors[] = new \WP_Error( 'pmp_im_member', sprintf( __( 'Could not decode timezone value (%s)', 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_timestamp'] ) );
						}
					}
					
					$order->updateTimeStamp(
						date( "Y", $timestamp ),
						date( "m", $timestamp ),
						date( "d", $timestamp ),
						date( "H:i:s", $timestamp )
					);
				}
			}
			
			// Add any Discount Code use for this user
			if ( ! empty( $this->pmpro_fields['membership_code_id'] ) && ! empty( $order ) && ! empty( $order->id ) ) {
				
				if ( false === $wpdb->insert(
						$pmpro_dc_uses_table,
						array(
							'code_id'   => $this->pmpro_fields['membership_code_id'],
							'user_id'   => $user_id,
							'order_id'  => $order->id,
							'timestamp' => 'CURRENT_TIMESTAMP',
						),
						array( '%d', '%d', '%d', '%s' )
					) ) {
					$errors[] = new \WP_Error( 'pmp_im_member', sprintf( __( 'Unable to set update discount code usage for code (ID: %d ) for user (user/order id: %d/%s)', 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_code_id'], $user_id, $order->id ) );
				}
			}
			
			if ( WP_DEBUG ) {
			    error_log("Should we send welcome email ({$this->send_welcome_email})? " . ( $this->send_welcome_email ? 'Yes' : 'No' ) );
            }
			
			// Email 'your membership account is active' to member if they were imported with an active member status
			if ( true == (bool) $this->send_welcome_email &&
			     ( isset( $this->pmpro_fields['membership_status'] ) && 'active' == $this->pmpro_fields['membership_status'] ) &&
			     1 === version_compare( PMPRO_VERSION, '1.9.5' )
			) {
                
			    $subject = null;
			    $body = null;
			    
				if ( ! empty( $pmproiufcsv_email ) ) {
					$subject = apply_filters( 'pmp_im_imported_member_message_subject', $pmproiufcsv_email['subject'] );
					$body    = apply_filters( 'pmp_im_imported_member_message_body', $pmproiufcsv_email['body'] );
				}
				
				global $pmproet_email_defaults;
				
				$template_name = 'imported_member';
				
				// Apply the saved
				if ( empty( $subject ) ) {
					
				    $subject = pmpro_getOption( "email_{$template_name}_subject" );
					
					if ( empty( $subject ) ) {
						$subject = isset( $pmproet_email_defaults[ $template_name ]['subject'] ) ? $pmproet_email_defaults[ $template_name ]['subject'] : __( "Your membership to !!sitename!! has been activated", 'pmpro-import-members-from-csv' );
					}
				}
				
				if ( WP_DEBUG ) {
					error_log("Using {$template_name} template for '{$subject}' message");
				}
				
				$email           = new \PMProEmail();
				$email->email    = $user->user_email;
				$email->data     = apply_filters( 'pmp_im_imported_member_message_data', array() );
				$email->subject  = $subject;
				$email->template = $template_name;
				
				if ( ! empty( $body ) ) {
				    $email->body = $body;
                } else {
					$email->body     = $this->load_email_body( null, $email->template );
                }
                
				$email->body = apply_filters( 'pmp_im_imported_member_message_body', $email->body );
				
				// Process and send email
				$email->sendEmail();
			}
			
			// Log errors to log file
			if ( ! empty( $errors ) ) {
				$this->log_errors( $errors );
			}
			
			// Update the error status
			if ( true === $has_error ) {
				update_option( 'e20r_import_errors', $has_error );
			}
		}
		
		if ( is_multisite() ) {
			switch_to_blog( $current_blog_id );
		}
	}
    
	/**
	 * Run data checks/validations (business rules, etc)
	 *
	 * @param int $user_id
	 *
	 * @return bool
	 *
	 * @access private
	 *
	 * @since  2.22 - ENHANCEMENT: Added validate_membership_data() method
	 * @since  2.30 - ENHANCEMENT: Multiple updates for payment gateway integrations, etc
	 */
	private function validate_membership_data( $user_id ) {
		
		$has_error = false;
		$errors    = array();
		
		// Make sure the membership_id column is present in the import file
		if ( false === is_numeric( $this->pmpro_fields['membership_id'] ) ) {
			$msg = sprintf( __( 'The membership ID (membership_id) column does not contain a numeric membership Level for user (ID: %d). (Membership data not imported!)', 'pmpro-import-members-from-csv' ), $user_id );
			
			$errors[]  = new \WP_Error( 'pmp_im_member', $msg );
			$has_error = true;
		}
		
		// Make sure the specified level ID exists on the system
		if ( isset( $this->pmpro_fields['membership_id'] ) && false === pmpro_getLevel( $this->pmpro_fields['membership_id'] ) ) {
			$msg = sprintf( __( 'The provided membership ID (%d) is not valid. (Membership data not imported for user (ID: %d)!)', 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_id'], $user_id );
			
			$errors[]  = new \WP_Error( 'pmp_im_member', $msg );
			$has_error = true;
		}
		
		/**
		 * Can't import membership if user has an invalid (non-existing) membership ID
		 */
		if ( false === $has_error && ( is_numeric( $this->pmpro_fields['membership_id'] ) && 0 < $this->pmpro_fields['membership_id'] ) ) {
			
			$found_level = pmpro_getLevel( $this->pmpro_fields['membership_id'] );
			
			if ( empty( $found_level ) ) {
				$msg = sprintf( __( "The membership ID (%d) specified for this user (ID: %d) is not a defined membership level, so we can't assign it. (Membership data not imported!)", 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_id'], $user_id );
				
				$errors[]  = new \WP_Error( 'pmp_im_member', $msg );
				$has_error = true;
			}
		}
		
		/**
		 * Not really an error, but we should warn about it!
		 * Warn if the user has an enddate in the future but have had their membership_status set to 'inactive'
		 */
		if ( ! empty( $this->pmpro_fields['membership_enddate'] ) &&
		     'inactive' == $this->pmpro_fields['membership_status'] &&
		     ( current_time( 'timestamp' ) < strtotime( $this->pmpro_fields['membership_enddate'], current_time( 'timestamp' ) ) )
		) {
			
			$msg      = sprintf( __( 'Warning: The membership (id: %d) for user ID %d will end at a future date (membership_enddate = %s), but you set the status to \'inactive\'...', 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_id'], $user_id, $this->pmpro_fields['membership_enddate'] );
			$errors[] = new \WP_Error( 'pmp_im_member', $msg );
		}
		
		/**
		 * Can't import membership if user has incorrect startdate format
		 */
		if ( ! empty( $this->pmpro_fields['membership_startdate'] ) &&
		     false === $this->validate_date( $this->pmpro_fields['membership_startdate'] ) ) {
			
			$msg       = sprintf( __( 'The membership start date (membership_startdate) column in the import .CSV file must use a MySQL formatted date (YYYY-MM-DD HH:MM:SS). You used \'%s\' (Membership data not imported!)', 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_startdate'] );
			$errors[]  = new \WP_Error( 'pmp_im_member', $msg );
			$has_error = true;
			
			$should_be = $this->convert_time( $this->pmpro_fields['membership_startdate'] );
			$should_be = false === $should_be ? current_time( 'timestamp' ) : $should_be;
			
			$this->display_errors['startdate'] = sprintf( __( 'The %2$smembership_startdate column%3$s contains an unrecognized date/time format. (Your format: \'%1$s\'. Expected format: \'%4$s\'). Membership data may not have been imported!', 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_startdate'], '<strong>', '</strong>', date( 'Y-m-d h:i:s', $should_be ) );
		}
		
		/**
		 * Can't import membership if user has incorrect enddate format
		 */
		if ( ! empty( $this->pmpro_fields['membership_enddate'] ) &&
		     false === $this->validate_date( $this->pmpro_fields['membership_enddate'] ) ) {
			
			$msg = sprintf( __( 'The membership end date (membership_enddate) column for the user (ID: %d) in the import .CSV file must use a MySQL formatted date (YYYY-MM-DD HH:MM:SS). You appear to have used \'%s\' (Membership data not imported!)', 'pmpro-import-members-from-csv' ), $user_id, $this->pmpro_fields['membership_enddate'] );
			
			$errors[]  = new \WP_Error( 'pmp_im_member', $msg );
			$has_error = true;
			
			$should_be = $this->convert_time( $this->pmpro_fields['membership_enddate'] );
			$should_be = false === $should_be ? current_time( 'timestamp' ) : $should_be;
			
			$this->display_errors['enddate'] = sprintf( __( 'The %2$smembership_enddate column%3$s contains an unrecognized date/time format. (Your format: \'%1$s\'. Expected format: \'%4$s\') Membership data may not have been imported!', 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_enddate'], '<strong>', '</strong>', date( 'Y-m-d h:i:s', $should_be ) );
		}
		
		if ( ! empty( $this->pmpro_fields['membership_status'] ) &&
		     ! in_array( $this->pmpro_fields['membership_status'], array( 'active', 'inactive', true ) ) ) {
			
			$msg = sprintf( __( "The membership_status column for user (ID: %d) contains an unexpected value (expected values: 'active' or 'inactive'). You used '%s' (Membership data not imported!)", 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_status'], $user_id );
			
			$errors[]  = new \WP_Error( 'pmp_im_member', $msg );
			$has_error = true;
		}
		
		/**
		 * Can't import if the cycle period is defined but has no cycle number
		 */
		if ( empty( $this->pmpro_fields['membership_cycle_number'] ) &&
		     ! empty( $this->pmpro_fields['membership_cycle_period'] )
		) {
			$msg      = sprintf( __( "Warning: You have configured a Membership Cycle Period (%s) for the user (ID %d), but you haven't included the number of periods ('membership_cycle_number')", 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_cycle_period'], $user_id );
			$errors[] = new \WP_Error( 'pmp_im_member', $msg );
		}
		
		/**
		 * Has a subscription_transaction_id, but we also expect there to be a payment transaction ID
		 */
		if ( ! empty( $this->pmpro_fields['membership_subscription_transaction_id'] ) && empty( $this->pmpro_fields['membership_payment_transaction_id'] ) ) {
			$msg      = sprintf( __( 'Warning: You have defined a subscription_transaction_id (%s) without also including a payment_transaction_id.', 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_subscription_transaction_id'] );
			$errors[] = new \WP_Error( 'pmp_im_member', $msg );
		}
		
		/**
		 * No gateway specified when importing subscription_transaction_id (or subscription transaction ID(s) without
		 */
		if ( ! empty( $this->pmpro_fields['membership_subscription_transaction_id'] ) && empty( $this->pmpro_fields['membership_gateway'] ) ) {
			$msg       = sprintf( __( "You need to specify both the membership_gateway (value: %s) and the membership_subscription_transaction_id (value: %s) to link active subscription plan(s) to order(s). (Skipping!)", 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_gateway'], $this->pmpro_fields['membership_subscription_transaction_id'] );
			$errors[]  = new \WP_Error( 'pmp_im_member', $msg );
			$has_error = true;
		}
		
		/**
		 * No gateway environment specified when importing a gateway integration
		 */
		if ( ! empty( $this->pmpro_fields['membership_gateway'] ) && empty( $this->pmpro_fields['membership_gateway_environment'] ) ) {
			$msg = sprintf( __( "Warning: You specified a membership_gateway (%s), but you didn't specify the gateway environment to use (membership_gateway_environment). Using current setting.", 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_gateway'] );
			
			$this->pmpro_fields['membership_gateway_environment'] = pmpro_getOption( 'gateway_environment' );
			$errors[]                                             = new \WP_Error( 'pmp_im_member', $msg );
			
		}
		
		/**
		 * Specified a gateway integration, but the value given is not what we'd expect
		 */
		if ( ! empty( $this->pmpro_fields['membership_gateway'] ) && ! empty( $this->pmpro_fields['membership_gateway_environment'] ) && ! in_array( $this->pmpro_fields['membership_gateway_environment'], array(
				'live',
				'sandbox',
			) ) ) {
			$msg       = sprintf( __( "You specified a payment gateway integration (membership_gateway) to use, but we do not recognize the gateway environment you have specified for this record (%s). (Skipping!)", 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_gateway_environment'] );
			$errors[]  = new \WP_Error( 'pmp_im_member', $msg );
			$has_error = true;
		}
		
		/**
		 * Did they specify a supported membership gateway?
		 */
		if ( ! empty( $this->pmpro_fields['membership_gateway'] ) ) {
			$gateways = pmpro_gateways();
			
			if ( ! in_array( $this->pmpro_fields['membership_gateway'], array_keys( $gateways ) ) ) {
				$msg                                      = sprintf( __( 'Warning: The payment gateway integration provided (membership_gateway: %s) is not one of the supported payment gateway integrations! (Using the default)', 'pmpro-import-members-from-csv' ), $this->pmpro_fields['membership_gateway'] );
				$errors[]                                 = new \WP_Error( 'pmp_im_member', $msg );
				$this->pmpro_fields['membership_gateway'] = pmpro_getOption( 'gateway' );
			}
		}
		
		/**
		 * Technically _NOT_ an error, but worth mentioning!
		 *
		 * membership_enddate + recurring billing data is probably _not_ correct, but we'll allow it.
		 */
		if ( ! empty( $this->pmpro_fields['membership_enddate'] ) &&
		     ! empty( $this->pmpro_fields['membership_cycle_number'] ) &&
		     ! empty( $this->pmpro_fields['membership_cycle_period'] ) ) {
			
			$msg      = sprintf( __( "Warning: You have an end date AND a recurring billing configuration (the membership_enddate, membership_billing_amount, membership_cycle_number and membership_cycle_period columns are not empty) user (ID: %d). Could result in an incorrectly configured membership for this user", 'pmpro-import-members-from-csv' ), $user_id );
			$errors[] = new \WP_Error( 'pmp_im_member', $msg );
		}
		
		// Save any error messages for the user
		if ( ! empty( $errors ) ) {
			$this->log_errors( $errors );
		}
		
		return apply_filters( 'pmpro_import_members_from_csv_data_check_status', $has_error, $user_id, $this->pmpro_fields );
	}
	
	/**
	 * @param string $billing_field_name
	 *
	 * @return string|bool
	 */
	private function map_billing_field_to_meta( $billing_field_name ) {
		
		$billing_fields = apply_filters( 'pmpro_import_members_billing_field_map', array(
				'billing_street'  => 'pmpro_baddress1',
				'billing_city'    => 'pmpro_bcity',
				'billing_state'   => 'pmpro_bstate',
				'billing_zip'     => 'pmpro_bzipcode',
				'billing_country' => 'pmpro_bcountry',
				'billing_phone'   => 'pmpro_bphone',
			)
		);
		
		if ( ! in_array( $billing_field_name, array_keys( $billing_fields ) ) ) {
			return null;
		}
		
		return $billing_fields[ $billing_field_name ];
	}
	
	/**
     * Load the body of the template
     *
     * @param string $body
	 * @param string $template_name
	 *
	 * @return string
     *
     * @since v2.50 - ENHANCEMENT: Include imported_member template in the  Email Template Admin add-on
	 */
	private function load_email_body( $body = null, $template_name ) {
		
		if ( WP_DEBUG ) {
			error_log( "Loading template text for {$template_name}" );
		}
		
		if ( !empty( $body ) ) {
		    return $body;
        }
        
		$email_text = '';
		
		global $pmproet_email_defaults;
		
		// Email disabled?
		if ( true === (bool) pmpro_getOption( "email_{$template_name}_disabled}" ) ) {
			return false;
		}
		
		// Not in the list of templates?
		if ( empty( $pmproet_email_defaults[ $template_name ] ) ) {
			return false;
		}
		
		if ( WP_DEBUG ) {
			error_log( "Setting the message subject for {$template_name} template" );
		}
		
		$template_body    = pmpro_getOption( "email_{$template_name}_body" );
		$template_header  = pmpro_getOption( 'email_header_body' );
		$template_footer  = pmpro_getOption( 'email_footer_body' );
		
		// Header disabled?
		if ( true !== (bool) pmpro_getOption( 'email_header_disabled' ) ) {
			
			if ( ! empty( $template_header ) ) {
				
				$email_text = $template_header;
				
			} else {
				
				$email_text = $this->load_template_part( 'header' );
			}
		}
		
		/**
		 * Have body content?
		 */
		if ( ! empty( $template_body ) ) {
			
			$email_text .= $template_body;
			
		} else {
			// Load the body content from a HTML template (somewhere)
			$email_text .= $this->load_template_part( $template_name );
		}
		
		// Footer disabled?
		if ( true === (bool) pmpro_getOption( 'email_footer_disabled' ) ) {
			
			if ( ! empty( $template_footer ) ) {
				$email_text .= $template_footer;
			} else {
				$email_text .= $this->load_template_part( 'footer' );
			}
		}
		
		return $email_text;
	}
	
	/**
	 * Load the body for the email message to send to the updated/new member
	 *
	 * @param string|null $template
	 *
	 * @return string|null
     *
     * @since v2.50 - ENHANCEMENT: Include imported_member template in the  Email Template Admin add-on
	 */
	private function load_template_part( $template ) {
	 
		global $pmproet_email_defaults;
		
		$locale = apply_filters( "plugin_locale", get_locale(), "pmpro-import-members-from-csv" );
		$body = null;
		
		// Load template from PMPro Email Templates Admin add-on
		if ( isset($pmproet_email_defaults[ $template ]['body'] ) && ! empty( $pmproet_email_defaults[ $template ]['body'] ) ) {
			
			$body = $pmproet_email_defaults[ $template ]['body'];
			
		} else {
			
			$locations = array(
				get_stylesheet_directory() . "/pmpro-import-members-from-csv/{$locale}/{$template}.html",
				get_stylesheet_directory() . "/pmpro-import-members-from-csv/{$template}.html",
				get_template_directory() . "/pmpro-import-members-from-csv/{$locale}/{$template}.html",
				get_template_directory() . "/pmpro-import-members-from-csv/{$template}.html",
				plugin_dir_path( __FILE__ ) . "/languages/emails/{$locale}/{$template}.html",
				plugin_dir_path( __FILE__ ) . "/emails/${template}.html"
			);
			
			$locations = apply_filters( 'pmpro_import_members_template_locations', $locations, $template );
			
			// Try to locate the template file in the file system
			foreach ( $locations as $path ) {
				
				if ( true === file_exists( $path ) ) {
					$body = file_get_contents( $path );
					break;
				}
			}
		}
		
		return $body;
	}
	
	/**
	 * Load Imported Member template to Email Templates Admin (add-on)
     *
     * @since v2.50 - ENHANCEMENT: Allow editing imported_member.html in the Email Templates Admin add-on
	 */
	public function add_email_templates() {
	    
	    global $pmproet_email_defaults;
	    
	    if ( WP_DEBUG ) {
	        error_log("Attempting to load template for the Welcome Imported Member message");
        }
        
		$pmproet_email_defaults['imported_member'] = array(
            'subject' => __( 'Welcome to my new website', 'pmpro-import-members-from-csv' ),
            'description' => __( 'Import: Welcome Member', 'pmpro-import-members-from-csv' ),
            'body' => file_get_contents( plugin_dir_path( __FILE__ ) . '/emails/imported_member.html' )
        );
    }
	
    /**
	 * Add links to support & docs for the plugin
	 *
	 * @param array  $links - Links for the Plugins page
	 * @param string $file
	 *
	 * @return array
	 */
	public function plugin_row_meta( $links, $file ) {
		
		if ( false !== strpos( $file, 'class.pmpro-import-members.php' ) ) {
			
			// Add (new) 'Import Users from CSV' links to plugin listing
			$new_links = array(
				'donate'        => sprintf(
					'<a href="%1$s" title="%2$s">%3$s</a>',
					esc_url_raw( 'https://www.paypal.me/eighty20results' ),
					__( 'Donate to support updates, maintenance and tech support for this plugin', 'pmpro' ),
					__( 'Donate', 'pmpro-import-members-from-csv' )
				),
				'documentation' => sprintf(
					'<a href="%1$s" title="%2$s">%3$s</a>',
					esc_url( 'https://wordpress.org/plugins/pmpro-import-members-from-csv/' ),
					__( 'View the documentation', 'pmpro-import-members-from-csv' ),
					__( 'Docs', 'pmpro-import-members-from-csv' )
				),
				'help'          => sprintf(
					'<a href="%1$s" title="%2$s">%3$s</a>',
					esc_url( 'https://wordpress.org/support/plugin/pmpro-import-members-from-csv' ),
					__( 'Visit the support forum', 'pmpro' ),
					__( 'Support', 'pmpro-import-members-from-csv' )
				),
			);
			
			$links = array_merge( $links, $new_links );
		}
		
		return $links;
	}
}

register_deactivation_hook( __FILE__, 'PMPRO\Addons\Import_Members_From_CSV::deactivation' );

// Load the plugin.
add_action( 'plugins_loaded', array( Import_Members_From_CSV::get_instance(), 'load_plugins' ) );
